from typing import Optional, List, Union
from hedra.versioning.flags.types.unstable.flag import unstable
from .distribution import Distribution
from .mutations.types.base.mutation import Mutation


@unstable
class Variant:

    def __init__(
        self,
        stage_name: str,
        weight: Optional[float] = None,
        distribution: str=None,
        distribution_intervals: int=10,
        mutations: Optional[List[Mutation]]=None
    ) -> None:
        self.stage_name = stage_name
        self.weight = weight
        self.distribution: Optional[Distribution] = None
        self.intervals = distribution_intervals
        self.mutations: Union[List[Mutation], None] = mutations
        
        if distribution:
            self.distribution = Distribution(
                distribution,
                intervals=distribution_intervals
            )

    def get_mutations(self) -> List[Mutation]:
        return [
            mutation.copy() for mutation in self.mutations
        ]



from .graphql.graphql_serializer import GraphQLSerializer
from .graphql_http2.graphql_http2_serializer import GraphQLHTTP2Serializer
from .grpc.grpc_serializer import GRPCSerializer
from .http.http_serializer import HTTPSerializer
from .http2.http2_serializer import HTTP2Serializer
from .http3.http3_serializer import HTTP3Serializer
from .playwright.playwright_serializer import PlaywrightSerializer
from .task.task_serializer import TaskSerializer
from .udp.udp_serializer import UDPSerializer
from .websocket.websocket_serializer import WebsocketSerializer


from __future__ import annotations
from hedra.distributed.models.http import (
    Request,
    Response
)
from pydantic import BaseModel
from typing import (
    Callable, 
    Union, 
    Dict, 
    Optional,
    List,
    Literal,
    Tuple
)
from .bidirectional_wrapper import BidirectionalWrapper
from .call_wrapper import CallWrapper
from .unidirectional_wrapper import UnidirectionalWrapper
from .types import MiddlewareType


class Middleware:

    def __init__(
        self,
        name: str,
        middleware_type: MiddlewareType=MiddlewareType.UNIDIRECTIONAL_BEFORE,
        methods: Optional[
            List[
                Literal[
                    "GET",
                    "HEAD",
                    "OPTIONS",
                    "POST",
                    "PUT",
                    "PATCH",
                    "DELETE",
                    "TRACE"
                ]
            ]
        ]=None,
        response_headers: Dict[str, str]={}
    ) -> None:
        
        self.name = name
        self.methods = methods
        self.response_headers = response_headers
        self.middleware_type = middleware_type
        self.wraps = False

        self._wrapper_types = {
            MiddlewareType.BIDIRECTIONAL: BidirectionalWrapper,
            MiddlewareType.CALL: CallWrapper,
            MiddlewareType.UNIDIRECTIONAL_BEFORE: UnidirectionalWrapper,
            MiddlewareType.UNIDIRECTIONAL_AFTER: UnidirectionalWrapper,
        }

    def __call__(self, request: Request) -> Tuple[
        Tuple[Response, int],
        bool
    ]:
        raise NotImplementedError('Err. __call__() should not be called on base Middleware class.')

    def wrap(
        self,
        handler: Callable[
            [Request],
            Union[
                BaseModel,
                str,
                None
            ]
        ]
    ):
        
        wrapper = self._wrapper_types.get(
            self.middleware_type,
            BidirectionalWrapper(
                self.name,
                handler,
                methods=self.methods,
                response_headers=self.response_headers,
                middleware_type=self.middleware_type 
            )
        )(
            self.name,
            handler,
            methods=self.methods,
            response_headers=self.response_headers,
            middleware_type=self.middleware_type 
        )

        if isinstance(wrapper, BidirectionalWrapper):
            wrapper.pre = self.__pre__
            wrapper.post = self.__post__

        elif isinstance(wrapper, (CallWrapper, UnidirectionalWrapper)):

            wrapper.run = self.__run__
        
            self.response_headers.update(wrapper.response_headers)

        wrapper.setup = self.__setup__
        self.wraps = wrapper.wraps
        
        return wrapper
    
    async def __setup__(self):
        pass
    
    async def __pre__(
        self, 
        request: Request,
        response: Response,
        status: int
    ) -> Tuple[
        Tuple[
            Request,
            Response, 
            int
        ],
        bool
    ]:
        raise NotImplementedError('Err. - __pre__() is not implemented for base Middleware class.')
    
    async def __post__(
        self, 
        request: Request,
        response: Response,
        status: int
    ) -> Tuple[
        Tuple[
            Request,
            Response, 
            int
        ],
        bool
    ]:
        raise NotImplementedError('Err. - __post__() is not implemented for base Middleware class.')
    
    async def __run__(
        self, 
        request: Request,
        response: Response,
        status: int
    ) -> Tuple[
        Tuple[Response, int],
        bool
    ]:
        raise NotImplementedError('Err. - __post__() is not implemented for base Middleware class.')
    
    async def run(
        self,
        request: Request
    ):
        raise NotImplementedError('Err. - middleware() is not implemented for base Middleware class.')
    


from hedra.core.graphs.stages.base.parallel.stage_priority import StagePriority
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.hooks.types.internal.decorator import Internal
from hedra.core.graphs.stages.base.stage import Stage


class Idle(Stage):
    stage_type=StageTypes.IDLE

    def __init__(self) -> None:
        super().__init__()
        self.name = self.__class__.__name__
        self.priority = None
        self.priority_level: StagePriority = StagePriority.map(
            self.priority
        )

        self.retries: int = 0

    @Internal()
    async def run(self):
        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Starting graph execution')

        



import json
import uuid
from datetime import datetime
from hedra.core.engines.types.websocket.action import WebsocketAction
from hedra.core.engines.types.common.types import RequestTypes
from typing import Dict, Any, List


try:
    from cassandra.cqlengine import columns
    from cassandra.cqlengine.models import Model
    has_connector = True

except Exception:
    columns = object
    Model = None
    has_connector = False


class CassandraWebsocketActionSchema:

    def __init__(self, table_name: str) -> None:
        self.action_columns = {
            'id': columns.UUID(primary_key=True, default=uuid.uuid4),
            'name': columns.Text(min_length=1, index=True),
            'method': columns.Text(min_length=1),
            'url': columns.Text(min_length=1),
            'headers': columns.Text(),
            'data': columns.Text(),
            'user': columns.Text(),
            'tags': columns.List(
                columns.Map(
                    columns.Text(),
                    columns.Text()
                )
            ),
            'created_at': columns.DateTime(default=datetime.now)
        }

        self.actions_table = type(
            f'{table_name}_websocket',
            (Model, ), 
            self.action_columns
        )

        self.type = RequestTypes.WEBSOCKET

        self._types_map: Dict[type, str] = {
            str: 'string',
            int: 'integer',
            float: 'float',
            bytes: 'bytes',
            bool: 'bool'
        }

        self._reverse_types_map: Dict[str, type] = {
            mapped_name: mapped_type for mapped_type, mapped_name in self._types_map.items()
        }

    def from_schema_record(
        self,
        action: Dict[str, Any]
    ) -> Dict[str, Any]:
        
        action_tags: List[Dict[str, Any]] = action.get('tags', [])
        action_data: Dict[str, str] = json.load(action.get('data'))

        data = action_data.get('data')
        datatype = action_data.get('datatype')

        if datatype == 'bytes':
            data = data.encode()

        elif datatype == 'integer':
            data = int(data)
        
        elif datatype == 'float':
            data = float(data)

        return {
            'engine': RequestTypes.GRAPHQL,
            'name': action.get('name'),
            'url': action.get('url'),
            'method': action.get('method'),
            'headers': json.loads(action.get('headers')),
            'data': data,
            'user': action.get('user'),
            'tags': [
                {
                    'name': tag.get('name'),
                    'value': self._reverse_types_map.get(
                        tag.get('datatype')
                    )(
                        tag.get('value')
                    )
                } for tag in action_tags
            ]
        }



import plotille
from collections import (
    OrderedDict,
    defaultdict
)
from tabulate import tabulate
from typing import (
    List, 
    Union,
    Dict
)
from hedra.reporting.system.system_metrics_set import SystemMetricsSet
from hedra.logging import HedraLogger


class SystemSummaryTable:

    def __init__(self) -> None:
        self.system_metrics_summaries: List[SystemMetricsSet] = []
        self.logger = HedraLogger()
        self.logger.initialize()

        self.cpu_table: Union[str, None] = None
        self.memory_table: Union[str, None] = None
        self.mb_per_vu_table: Union[str, None] = None
        self.graph_cpu_table: Union[str, None] = None
        self.graph_memory_table: Union[str, None] = None

        self.cpu_plot_rows: Dict[str, List[Union[int, float]]] = defaultdict(list)
        self.memory_plot_rows: Dict[str, List[Union[int, float]]] = defaultdict(list)
        self.graph_metrics_summary: Union[SystemMetricsSet, None] = None

        self.enabled_tables = {
            'system': False
        }

    def generate_tables(self):
        self.graph_cpu_table = self._to_graph_cpu_table()
        self.graph_memory_table = self._to_graph_memory_table()
        
        self.cpu_table = self._to_cpu_table()
        self.memory_table = self._to_memory_table()
        self.mb_per_vu_table = self._to_mb_per_vu_table()

    def show_tables(self):
        
        if any(self.enabled_tables.values()):
            self.logger.console.sync.info('\n-- System Metrics --')

        if self.enabled_tables.get('system'):
            for stage_name, monitor in self.graph_metrics_summary.cpu.stage_metrics.items():
                for monitor_name, metrics in monitor.items():

                    show_plot = self.graph_metrics_summary.cpu.visibility_filters[stage_name][monitor_name]

                    if show_plot:
                        scatter_plot = plotille.scatter(
                            [idx for idx in range(
                                1, 
                                len(metrics) + 1
                            )],
                            metrics,
                            width=120,
                            height=10,
                            y_min=0,
                            x_min=0,
                            x_max=len(metrics) + 1,
                            linesep='\n',
                            X_label='time (sec)',
                            Y_label='pct. utilization per thread',
                            lc='cyan',
                            marker='⨯'
                        )

                        self.logger.console.sync.info(f'''\n{monitor_name} % CPU Usage (Per Worker)\n''')
                        self.logger.console.sync.info(f'''{scatter_plot}\n''')

            for stage_name, monitor in self.graph_metrics_summary.memory.stage_metrics.items():
                for monitor_name, metrics in monitor.items():
                    
                    show_plot = self.graph_metrics_summary.memory.visibility_filters[stage_name][monitor_name]

                    if show_plot:
                        scatter_plot = plotille.scatter(
                            [idx for idx in range(
                                1, 
                                len(metrics) + 1
                            )],
                            [
                                round(
                                    metric_value/(1024**3),
                                    2
                                ) for metric_value in metrics
                            ],
                            width=120,
                            height=10,
                            y_min=0,
                            x_min=0,
                            x_max=len(metrics) + 1,
                            linesep='\n',
                            X_label='time (sec)',
                            Y_label='memory used (gb)',
                            lc='cyan',
                            marker='⨯'
                        )

                        self.logger.console.sync.info(f'''\n{monitor_name} % Memory Usage (gb)\n''')
                        self.logger.console.sync.info(f'''{scatter_plot}\n''')

            self.logger.console.sync.info('\nCPU (% per worker):\n')
            self.logger.console.sync.info(f'''{self.graph_cpu_table}\n''')

            self.logger.console.sync.info('\nMemory (gb):\n')
            self.logger.console.sync.info(f'''{self.graph_memory_table}\n''')

        if self.enabled_tables.get('system') and self.enabled_tables.get('stages'):
            
            seen_plots: List[str] = []

            for metrics_set in self.system_metrics_summaries:
                for stage_name, monitor in metrics_set.cpu.stage_metrics.items():
                    for monitor_name, metrics in monitor.items():

                        show_plot = metrics_set.cpu.visibility_filters[stage_name][monitor_name]

                        if show_plot and monitor_name not in seen_plots:
                            scatter_plot = plotille.scatter(
                                [idx for idx in range(
                                    1, 
                                    len(metrics) + 1
                                )],
                                metrics,
                                width=120,
                                height=10,
                                y_min=0,
                                x_min=0,
                                x_max=len(metrics) + 1,
                                linesep='\n',
                                X_label='time (sec)',
                                Y_label='pct. utilization per thread',
                                lc='cyan',
                                marker='⨯'
                            )

                            self.logger.console.sync.info(f'''\n{monitor_name} % CPU Usage (Per Worker)\n''')
                            self.logger.console.sync.info(f'''{scatter_plot}\n''')

                            seen_plots.append(monitor_name)

            seen_plots: List[str] = []

            for metrics_set in self.system_metrics_summaries:
                for stage_name, monitor in metrics_set.memory.stage_metrics.items():
                    for monitor_name, metrics in monitor.items():
                        
                        show_plot = metrics_set.cpu.visibility_filters[stage_name][monitor_name]

                        if show_plot and monitor_name not in seen_plots:
                            scatter_plot = plotille.scatter(
                                [idx for idx in range(
                                    1, 
                                    len(metrics) + 1
                                )],
                                [
                                    round(
                                        metric_value/(1024**3),
                                        2
                                    ) for metric_value in metrics
                                ],
                                width=120,
                                height=10,
                                y_min=0,
                                x_min=0,
                                x_max=len(metrics) + 1,
                                linesep='\n',
                                X_label='time (sec)',
                                Y_label='memory used (gb)',
                                lc='cyan',
                                marker='⨯'
                            )

                            self.logger.console.sync.info(f'''\n{monitor_name} % Memory Usage (gb)\n''')
                            self.logger.console.sync.info(f'''{scatter_plot}\n''')

                            seen_plots.append(monitor_name)

            self.logger.console.sync.info('\nCPU (% per worker):\n')
            self.logger.console.sync.info(f'''{self.cpu_table}\n''')

            self.logger.console.sync.info('\nMemory (gb):\n')
            self.logger.console.sync.info(f'''{self.memory_table}\n''')

            self.logger.console.sync.info('\nMemory per VU (mb):\n')
            self.logger.console.sync.info(f'''{self.mb_per_vu_table}\n''')

    def _to_graph_cpu_table(self):

        table_row = OrderedDict()
        for metric_group in self.graph_metrics_summary.cpu:
            for row_name in SystemMetricsSet.metrics_table_keys:
                    table_row[row_name] = metric_group.record.get(row_name)

        return tabulate(
            [table_row],
            headers='keys',
            missingval='None',
            tablefmt="simple",
            floatfmt='.2f'
        )
    
    def _to_graph_memory_table(self):

        table_row = OrderedDict()
        for metric_group in self.graph_metrics_summary.memory:
            for row_name in SystemMetricsSet.metrics_table_keys:
                    table_row[row_name] = metric_group.record.get(row_name)

        return tabulate(
            [table_row],
            headers='keys',
            missingval='None',
            tablefmt="simple",
            floatfmt='.2f'
        )

    def _to_cpu_table(self):
        
        seen_monitors: List[str] = []
        table_rows: List[OrderedDict] = []

        for metrics_set in self.system_metrics_summaries:
            for metric_group in metrics_set.cpu:
                
                if metric_group.name not in seen_monitors:

                    table_row = OrderedDict()

                    for row_name in SystemMetricsSet.metrics_table_keys:
                            table_row[row_name] = metric_group.record.get(row_name)

                    table_rows.append(table_row)
                    seen_monitors.append(metric_group.name)

        return tabulate(
            list(sorted(
                table_rows,
                key=lambda row: row['name']
            )),
            headers='keys',
            missingval='None',
            tablefmt="simple",
            floatfmt='.2f'
        )
    
    def _to_memory_table(self):

        seen_monitors: List[str] = []
        table_rows: List[OrderedDict] = []

        for metrics_set in self.system_metrics_summaries:
            for stage_metrics in metrics_set.memory.metrics.values():
                for metric_group in stage_metrics.values():
                    
                    if metric_group.name not in seen_monitors:

                        table_row = OrderedDict()

                        for row_name in SystemMetricsSet.metrics_table_keys:
                                table_row[row_name] = metric_group.record.get(row_name)

                        table_rows.append(table_row)
                        seen_monitors.append(metric_group.name)

        return tabulate(
            list(sorted(
                table_rows,
                key=lambda row: row['name']
            )),
            headers='keys',
            missingval='None',
            tablefmt="simple",
            floatfmt='.2f'
        )
    
    def _to_mb_per_vu_table(self):

        seen_monitors: List[str] = []
        table_rows: List[OrderedDict] = []

        for metrics_set in self.system_metrics_summaries:
            for monitor_name, stage_metrics in metrics_set.mb_per_vu.items():

                if monitor_name not in seen_monitors:
                        
                    table_row = OrderedDict()

                    for row_name in SystemMetricsSet.metrics_table_keys:
                        table_row[row_name] = stage_metrics.record.get(row_name)

                    table_rows.append(table_row)
                    seen_monitors.append(monitor_name)

        return tabulate(
            list(sorted(
                table_rows,
                key=lambda row: row['name']
            )),
            headers='keys',
            missingval='None',
            tablefmt="simple",
            floatfmt='.2f'
        )




import time
from hedra.distributed.env import (
    load_env,
    ReplicationEnv
)
from hedra.distributed.env.time_parser import TimeParser
from hedra.distributed.models.raft.logs import Entry
from hedra.distributed.snowflake.snowflake_generator import Snowflake
from typing import List, Dict, Union
from .errors import InvalidTermError


class LogQueue:

    def __init__(self) -> None:

        env = load_env(ReplicationEnv)

        self.logs: List[Entry] = []
        self._timestamps: List[float] = []
        self._commits: List[float] = []
        self.timestamp_index_map: Dict[float, int] = {}
        self._term = 0
        self.size = 0
        self.commit_index = 0
        self._last_timestamp = 0
        self._last_commit_timestamp = 0
        self._prune_max_age = TimeParser(
            env.MERCURY_SYNC_RAFT_LOGS_PRUNE_MAX_AGE
        ).time
        self._prune_max_count = env.MERCURY_SYNC_RAFT_LOGS_PRUNE_MAX_COUNT

    @property
    def last_timestamp(self):

        if len(self._timestamps) > 0:
            return self._timestamps[-1]
        
        else:
            return 0

    def latest(self):

        if len(self._commits) > 0:
            latest_commit_timestamp = self._commits[-1]
            latest_index = self.timestamp_index_map[latest_commit_timestamp]

        else:
            latest_index = 0

        return self.logs[latest_index:]
    
    def commit(self):

        if len(self._timestamps) > 0:
            self._last_commit_timestamp = self._timestamps[-1]
            self._commits.append(self._last_commit_timestamp)

    def get(self, shard_id: int):
        flake = Snowflake.parse(shard_id)

        index = self.timestamp_index_map.get(flake.timestamp, -1)

        if self.size < 1:
            return None

        return self.logs[index]
    
    def filter(self, key: str):
        return [
            entry for entry in self.logs if entry.key == key
        ]
    
    def update(
        self, 
        entries: List[Entry]
    ) -> Union[Exception, None]:

        last_entry = entries[-1]
        
        last_entry_id = Snowflake.parse(last_entry.entry_id)
        last_entry_term = last_entry.term

        if last_entry_term < self._term:
            return InvalidTermError(
                last_entry_id,
                last_entry_term,
                self._term
            )
        
        # Did we miss an election or havent caught on to a leader change? let's update!
        elif last_entry_term > self._term:
            self._term = last_entry_term

        if self.size < 1:

            for idx, entry in enumerate(entries):

                entry_id = Snowflake.parse(entry.entry_id)
                entry_timestamp = entry_id.timestamp

                self.timestamp_index_map[entry_timestamp] = idx
                self._timestamps.append(entry_timestamp)
                self.logs.append(entry)

                self.size += 1

        else:

            for entry in entries:

                if len(self._timestamps) > 0:
                    last_queue_timestamp = self._timestamps[-1]  

                else:
                    last_queue_timestamp = 0

                next_index = self.size

                entry_id = Snowflake.parse(entry.entry_id)
                entry_timestamp = entry_id.timestamp

                # We've received a missing entry so insert it in order..
                if entry_timestamp < last_queue_timestamp:

                    # The insert index is at the index of last timestamp less 
                    # than the entry timestamp + 1.
                    #
                    # I.e. if the last idx < timestamp is 4 we insert at 5.
                    #

                    previous_timestamps = [
                        idx for idx, timestamp in enumerate(self._timestamps) if timestamp < entry_timestamp
                    ]

                    if len(previous_timestamps) > 0:
                        
                        last_previous_timestamp_idx = previous_timestamps[-1]

                        insert_index: int = last_previous_timestamp_idx + 1

                        next_logs = self.logs[insert_index:]
                        next_timestamps = self._timestamps[insert_index:]

                        previous_logs = self.logs[:insert_index]
                        previous_timestamps = self._timestamps[:insert_index]

                    else:
                        
                        insert_index = 0

                        next_logs = self.logs
                        next_timestamps = self._timestamps

                        previous_logs = []
                        previous_timestamps = []
                
                    previous_logs.append(entry)
                    previous_timestamps.append(entry_timestamp)

                    previous_logs.extend(next_logs)
                    previous_timestamps.extend(next_timestamps)

                    self.timestamp_index_map[entry_timestamp] = insert_index

                    for timestamp in next_timestamps:
                        self.timestamp_index_map[timestamp] += 1

                    self.logs = previous_logs
                    self._timestamps = previous_timestamps

                    self.size += 1
                
                # We've received entries to append
                elif entry_timestamp > last_queue_timestamp:
                        
                    self.logs.append(entry)
                    self._timestamps.append(entry_timestamp)
                
                    self.timestamp_index_map[entry_timestamp] = next_index
                    self.size += 1

                # We've receive an entry to replace.
                else:
                    
                    next_index = self.timestamp_index_map[entry_timestamp]

                    self.logs[next_index] = entry   
                    self._timestamps[next_index] = entry_timestamp
            

    def prune(self):

        current_time = int(time.time() * 1000)
        
        # Get the number of timestamps older than our max prune age
        count = len([
            timestamp for timestamp in self._timestamps if current_time - timestamp > self._prune_max_age
        ])

        # If greater than our max prune count, set prune count as max prune count.
        if count > self._prune_max_count:
            count = self._prune_max_count

        if count >= self.size:

            self.logs = []
            self._timestamps = []
            self.timestamp_index_map = {}
            self._commits = []

            self.size = 0
            self.commit_index = 0
            self._last_timestamp = 0
            self._last_commit_timestamp = 0
            self.size = 0

        else:

            pruned_timestamps = self._timestamps[:count]

            for timestamp in pruned_timestamps:
                if self.timestamp_index_map.get(timestamp):
                    del self.timestamp_index_map[timestamp]

            self.logs = self.logs[count:]
            self._timestamps = self._timestamps[count:]

            self._commits = [
                commit for commit in self._commits if commit > self._timestamps[0]
            ]

            self.size -= count
            

        




import signal
import asyncio
from asyncio import (
    Future,
    Transport,
    get_event_loop,
    AbstractEventLoop
)
from asyncio.exceptions import LimitOverrunError
from .constants import _DEFAULT_LIMIT


class Reader:

    _source_traceback = None
    __slots__ = (
        '_limit',
        '_loop',
        '_buffer',
        '_eof',
        '_waiter',
        '_exception',
        '_transport',
        '_paused',
        '__weakref__'
    )

    def __init__(self, limit=_DEFAULT_LIMIT, loop=None):
        # The line length limit is  a security feature;
        # it also doubles as half the buffer limit.

        if limit <= 0:
            raise ValueError('Limit cannot be <= 0')

        self._limit = limit
        if loop is None:
            self._loop = get_event_loop()
        else:
            self._loop: AbstractEventLoop = loop

        self._buffer = bytearray()
        self._eof = False    # Whether we're done.
        self._waiter: Future = None  # A future used by _wait_for_data()
        self._exception = None
        self._transport: Transport = None
        self._paused = False

    def exception(self):
        return self._exception

    def set_exception(self, exc):
        self._exception = exc

        waiter = self._waiter
        if waiter is not None:
            self._waiter = None
            if not waiter.cancelled():
                waiter.set_exception(exc)

    def set_transport(self, transport):
        # assert self._transport is None, 'Transport already set'
        self._transport = transport

    def _maybe_resume_transport(self):
        if self._paused and len(self._buffer) <= self._limit:
            self._paused = False
            self._transport.resume_reading()

    def feed_eof(self):
        self._eof = True

        waiter = self._waiter
        if waiter is not None:
            self._waiter = None
            if not waiter.cancelled():
                waiter.set_result(None)

    def at_eof(self):
        """Return True if the buffer is empty and 'feed_eof' was called."""
        return self._eof and not self._buffer

    def feed_data(self, data):
        assert not self._eof, 'feed_data after feed_eof'

        if not data:
            return

        self._buffer.extend(data)
        
        waiter = self._waiter
        if waiter is not None:
            self._waiter = None
            if not waiter.cancelled():
                waiter.set_result(None)

        if self._transport and self._paused == False and len(self._buffer) > 2 * self._limit:
            try:
                self._transport.pause_reading()
            except NotImplementedError:
                # The transport can't be paused.
                # We'll just have to buffer all data.
                # Forget the transport so we don't keep trying.
                self._transport = None
            else:
                self._paused = True

    async def _wait_for_data(self, func_name):
        """Wait until feed_data() or feed_eof() is called.
        If stream was paused, automatically resume it.
        """
        # StreamReader uses a future to link the protocol feed_data() method
        # to a read coroutine. Running two read coroutines at the same time
        # would have an unexpected behaviour. It would not possible to know
        # which coroutine would get the next data.
        if self._waiter:
            raise RuntimeError(
                f'{func_name}() called while another coroutine is '
                f'already waiting for incoming data'
            )

        assert not self._eof, '_wait_for_data after EOF'

        # Waiting for data while paused will make deadlock, so prevent it.
        # This is essential for readexactly(n) for case when n > self._limit.
        if self._paused:
            self._paused = False
            self._transport.resume_reading()

        self._waiter = self._loop.create_future()
        try:
            await self._waiter
        finally:
            self._waiter = None

    async def read(self, n=-1):

        if not self._buffer and not self._eof:
            await self._wait_for_data('read')

        data = bytes(self._buffer[:n])
        del self._buffer[:n]

        self._maybe_resume_transport()
        return data
    
    async def readline_fast(self, sep=b'\n'):
        seplen = len(sep)
        if self._exception is not None:
            raise self._exception
        
        if not self._buffer:

            if self._paused:
                self._paused = False
                self._transport.resume_reading()

            self._waiter = self._loop.create_future()
            try:
                await self._waiter
            finally:
                self._waiter = None

        isep = self._buffer.find(sep)
        if isep < 0:
            chunk = bytes(self._buffer)
            self._buffer.clear()
            return chunk


        chunk = self._buffer[:isep + seplen]
        del self._buffer[:isep + seplen]
        self._maybe_resume_transport()
        return bytes(chunk)

    async def readuntil(self, separator=b'\n'):
        """Read data from the stream until ``separator`` is found.
        On success, the data and separator will be removed from the
        internal buffer (consumed). Returned data will include the
        separator at the end.
        Configured stream limit is used to check result. Limit sets the
        maximal length of data that can be returned, not counting the
        separator.
        If an EOF occurs and the complete separator is still not found,
        an IncompleteReadError exception will be raised, and the internal
        buffer will be reset.  The IncompleteReadError.partial attribute
        may contain the separator partially.
        If the data cannot be read because of over limit, a
        LimitOverrunError exception  will be raised, and the data
        will be left in the internal buffer, so it can be read again.
        """
        seplen = len(separator)

        if self._exception is not None:
            raise self._exception

        # Consume whole buffer except last bytes, which length is
        # one less than seplen. Let's check corner cases with
        # separator='SEPARATOR':
        # * we have received almost complete separator (without last
        #   byte). i.e buffer='some textSEPARATO'. In this case we
        #   can safely consume len(separator) - 1 bytes.
        # * last byte of buffer is first byte of separator, i.e.
        #   buffer='abcdefghijklmnopqrS'. We may safely consume
        #   everything except that last byte, but this require to
        #   analyze bytes of buffer that match partial separator.
        #   This is slow and/or require FSM. For this case our
        #   implementation is not optimal, since require rescanning
        #   of data that is known to not belong to separator. In
        #   real world, separator will not be so long to notice
        #   performance problems. Even when reading MIME-encoded
        #   messages :)

        # `offset` is the number of bytes from the beginning of the buffer
        # where there is no occurrence of `separator`.
        offset = 0

        # Loop until we find `separator` in the buffer, exceed the buffer size,
        # or an EOF has happened.
        while True:
            buflen = len(self._buffer)

            # Check if we now have enough data in the buffer for `separator` to
            # fit.
            if buflen - offset >= seplen:
                isep = self._buffer.find(separator, offset)

                if isep != -1:
                    # `separator` is in the buffer. `isep` will be used later
                    # to retrieve the data.
                    break

                # see upper comment for explanation.
                offset = buflen + 1 - seplen
                if offset > self._limit:
                    raise LimitOverrunError(
                        'Separator is not found, and chunk exceed the limit',
                        offset)

            # Complete message (with full separator) may be present in buffer
            # even when EOF flag is set. This may happen when the last chunk
            # adds data which makes separator be found. That's why we check for
            # EOF *ater* inspecting the buffer.
            if self._eof:
                chunk = bytes(self._buffer)
                self._buffer.clear()
                raise Exception('Connection closed.')

            # _wait_for_data() will resume reading if stream was paused.
            if not self._buffer:
                await self._wait_for_data('readuntil')

        if isep > self._limit:
            raise LimitOverrunError(
                'Separator is found, but chunk is longer than limit', isep)

        chunk = self._buffer[:isep + seplen]
        del self._buffer[:isep + seplen]
        self._maybe_resume_transport()
        return bytes(chunk)

    async def read_headers(self, separator=b'\n'):
        """Read data from the stream until ``separator`` is found.
        On success, the data and separator will be removed from the
        internal buffer (consumed). Returned data will include the
        separator at the end.
        Configured stream limit is used to check result. Limit sets the
        maximal length of data that can be returned, not counting the
        separator.
        If an EOF occurs and the complete separator is still not found,
        an IncompleteReadError exception will be raised, and the internal
        buffer will be reset.  The IncompleteReadError.partial attribute
        may contain the separator partially.
        If the data cannot be read because of over limit, a
        LimitOverrunError exception  will be raised, and the data
        will be left in the internal buffer, so it can be read again.
        """
        headers = {}
        seplen = len(separator)

        while True:

            if self._exception is not None:
                raise self._exception
            
            if not self._buffer:

                if self._paused:
                    self._paused = False
                    self._transport.resume_reading()

                self._waiter = self._loop.create_future()
                try:
                    await self._waiter
                finally:
                    self._waiter = None

            isep = self._buffer.find(separator)
            if isep < 0:
                chunk = bytes(self._buffer)
                self._buffer.clear()

            else:
                chunk = bytes(self._buffer[:isep + seplen])
                del self._buffer[:isep + seplen]
                self._maybe_resume_transport()

            if b':' not in chunk:
                break
            
            decoded = chunk.strip().split(b':',1)

            key, value = decoded
            headers[bytes(key).lower()] = value.strip()
            
        return headers

    async def iter_headers(self, separator=b'\n'):
        """Read data from the stream until ``separator`` is found.
        On success, the data and separator will be removed from the
        internal buffer (consumed). Returned data will include the
        separator at the end.
        Configured stream limit is used to check result. Limit sets the
        maximal length of data that can be returned, not counting the
        separator.
        If an EOF occurs and the complete separator is still not found,
        an IncompleteReadError exception will be raised, and the internal
        buffer will be reset.  The IncompleteReadError.partial attribute
        may contain the separator partially.
        If the data cannot be read because of over limit, a
        LimitOverrunError exception  will be raised, and the data
        will be left in the internal buffer, so it can be read again.
        """
        headers = {}
        seplen = len(separator)

        while True:

            if self._exception is not None:
                raise self._exception
            
            if not self._buffer:

                if self._paused:
                    self._paused = False
                    self._transport.resume_reading()

                self._waiter = self._loop.create_future()
                try:
                    await self._waiter
                finally:
                    self._waiter = None

            isep = self._buffer.find(separator)
            if isep < 0:
                chunk = bytes(self._buffer)
                self._buffer.clear()

            else:
                chunk = bytes(self._buffer[:isep + seplen])
                del self._buffer[:isep + seplen]
                self._maybe_resume_transport()

            if b':' not in chunk:
                break
            
            decoded = chunk.strip().split(b':',1)

            key, value = decoded
            yield key.lower(), value, chunk

    async def __aiter__(self):
        line = await self.readuntil()
        yield line

    async def __anext__(self):
        val = await self.readuntil()
        if val == b'':
            raise StopAsyncIteration
        return val



from enum import Enum


class LoggerTypes(Enum):
    CONSOLE='console'
    DISTRIBUTED='distributed'
    FILESYSTEM='filesystem'
    DISTRIBUTED_FILESYSTEM='distributed_filesysem'
    HEDRA='hedra'
    SPINNER='spinner'





import functools
import os
from typing import Tuple, Union
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.registrar import registrar
from hedra.data.connectors.connector import (
    AWSLambdaConnectorConfig,
    BigTableConnectorConfig,
    CassandraConnectorConfig,
    CosmosDBConnectorConfig,
    CSVConnectorConfig,
    GoogleCloudStorageConnectorConfig,
    HARConnectorConfig,
    JSONConnectorConfig,
    KafkaConnectorConfig,
    MongoDBConnectorConfig,
    MySQLConnectorConfig,
    PostgresConnectorConfig,
    RedisConnectorConfig,
    S3ConnectorConfig,
    SnowflakeConnectorConfig,
    SQLiteConnectorConfig,
    XMLConnectorConfig
)
from .validator import LoadHookValidator


@registrar(HookType.LOAD)
def load(
    *names: Tuple[str, ...], 
    loader: Union[
        AWSLambdaConnectorConfig,
        BigTableConnectorConfig,
        CassandraConnectorConfig,
        CosmosDBConnectorConfig,
        CSVConnectorConfig,
        GoogleCloudStorageConnectorConfig,
        HARConnectorConfig,
        JSONConnectorConfig,
        KafkaConnectorConfig,
        MongoDBConnectorConfig,
        MySQLConnectorConfig,
        PostgresConnectorConfig,
        RedisConnectorConfig,
        S3ConnectorConfig,
        SnowflakeConnectorConfig,
        SQLiteConnectorConfig,
        XMLConnectorConfig
    ]=None, 
    order: int=1,
    skip: bool=False
):
    
    LoadHookValidator(
        names=names,
        loader=loader,
        order=order,
        skip=skip
    )
    
    def wrapper(func):

        @functools.wraps(func)
        def decorator(*args, **kwargs):
            return func(*args, **kwargs)

        return decorator

    return wrapper


from __future__ import annotations
from weakref import ref
from asyncio import (
    Transport,
    Protocol
)
from asyncio.coroutines import iscoroutine
from hedra.core.engines.types.common.protocols.shared import FlowControlMixin
from hedra.core.engines.types.common.protocols.shared.reader import Reader
from hedra.core.engines.types.common.protocols.shared.writer import Writer


class TCPProtocol(FlowControlMixin, Protocol):
    """Helper class to adapt between Protocol and StreamReader.
    (This is a helper class instead of making StreamReader itself a
    Protocol subclass, because the StreamReader has other potential
    uses, and to prevent the user of the StreamReader to accidentally
    call inappropriate methods of the protocol.)
    """

    __slots__ = (
        '_source_traceback',
        '_reject_connection',
        '_stream_writer',
        '_transport',
        '_client_connected_cb',
        '_over_ssl',
        '_closed',
        '_stream_reader_wr'
    )

    def __init__(self, stream_reader: Reader, client_connected_cb=None, loop=None):
        super().__init__(loop=loop)
        self._source_traceback = None

        if stream_reader is not None:
            self._stream_reader_wr: Reader = ref(stream_reader)
            self._source_traceback = stream_reader._source_traceback
        else:
            self._stream_reader_wr = None
        if client_connected_cb is not None:
            # This is a stream created by the `create_server()` function.
            # Keep a strong reference to the reader until a connection
            # is established.
            self._strong_reader = stream_reader
        self._reject_connection = False
        self._stream_writer: Writer = None
        self._transport: Transport = None
        self._client_connected_cb = client_connected_cb
        self._over_ssl = False
        self._closed = self._loop.create_future()

    @property
    def _stream_reader(self) -> Reader:
        if self._stream_reader_wr is None:
            return None
        return self._stream_reader_wr()

    def _replace_writer(self, writer: Writer):
        transport = writer.transport
        self._stream_writer = writer
        self._transport = transport
        self._over_ssl = transport.get_extra_info('sslcontext') is not None

    def connection_made(self, transport: Transport):
        if self._reject_connection:
            context = {
                'message': ('An open stream was garbage collected prior to '
                            'establishing network connection; '
                            'call "stream.close()" explicitly.')
            }
            if self._source_traceback:
                context['source_traceback'] = self._source_traceback
            self._loop.call_exception_handler(context)
            transport.abort()
            return
        self._transport = transport
        reader: Reader = self._stream_reader
        if reader is not None:
            reader.set_transport(transport)
        self._over_ssl = transport.get_extra_info('sslcontext') is not None
        if self._client_connected_cb is not None:
            self._stream_writer = Reader(transport, self,
                                               reader,
                                               self._loop)
            res = self._client_connected_cb(reader,
                                            self._stream_writer)
            if iscoroutine(res):
                self._loop.create_task(res)
            self._strong_reader = None

    def connection_lost(self, exc):
        reader: Reader = self._stream_reader
        if reader is not None:
            if exc is None:
                reader.feed_eof()
            else:
                reader.set_exception(exc)
        if not self._closed.done():
            if exc is None:
                self._closed.set_result(None)
            else:
                self._closed.set_exception(exc)
        super().connection_lost(exc)
        self._stream_reader_wr = None
        self._stream_writer = None
        self._transport = None

    def data_received(self, data):
        reader = self._stream_reader
        if reader is not None:
            reader.feed_data(data)

    def eof_received(self):
        reader: Reader = self._stream_reader
        if reader is not None:
            reader.feed_eof()
        if self._over_ssl:
            # Prevent a warning in SSLProtocol.eof_received:
            # "returning true from eof_received()
            # has no effect when using ssl"
            return False
        return True

    def _get_close_waiter(self, stream):
        return self._closed

    def __del__(self):
        # Prevent reports about unhandled exceptions.
        # Better than self._closed._log_traceback = False hack
        try:
            closed = self._closed
        except AttributeError:
            pass  # failed constructor
        else:
            if closed.done() and not closed.cancelled():
                closed.exception()




from typing import Optional
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt
)
from hedra.data.connectors.common.connector_type import ConnectorType

class SnowflakeConnectorConfig(BaseModel):
    username: StrictStr
    password: StrictStr
    organization_id: StrictStr
    account_id: StrictStr
    private_key: Optional[StrictStr]
    warehouse: StrictStr
    database: StrictStr
    database_schema: StrictStr='PUBLIC'
    table_name: StrictStr
    connect_timeout: StrictInt=30
    connector_type: ConnectorType=ConnectorType.Snowflake

    class Config:
        arbitrary_types_allowed = True


from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictFloat,
    StrictBool,
    conlist
)
from typing import (
    Union, 
    List, 
    Literal,
    Optional,
    Dict
)


class CorsHeaders(BaseModel):
    access_control_allow_origin: conlist(
        StrictStr,
        min_items=1
    )
    access_control_expose_headers: Optional[List[StrictStr]]
    access_control_max_age: Optional[Union[StrictInt, StrictFloat]]
    access_control_allow_credentials: Optional[StrictBool]
    access_control_allow_methods: conlist(
        Literal[
            "GET",
            "HEAD",
            "OPTIONS",
            "POST",
            "PUT",
            "PATCH",
            "DELETE",
            "TRACE"
        ],
        min_items=1
    )
    access_control_allow_headers: Optional[List[StrictStr]]

    def to_headers(self):

        cors_headers: Dict[str, str] = {}

        headers = self.dict(exclude_none=True)

        for key, value in headers.items():
            
            header_key = '-'.join([
                segment.capitalize() for segment in key.split('_')
            ])

            if key == 'access_control_allow_origin':
                header_value = ' | '.join(value)

            elif key == 'access_control_max_age':
                header_value = "true" if value else "false"

            else:
                header_value = ', '.join(value)

            cors_headers[header_key] = header_value

        return cors_headers
    
    def to_simple_headers(self):


        allow_all_origins = False        
        allow_all_origins = "*" in self.access_control_allow_origin
        simple_headers: Dict[str, str] = {}

        if allow_all_origins:
            simple_headers["Access-Control-Allow-Origin"] = "*"

        if self.access_control_allow_credentials:
            simple_headers["Access-Control-Allow-Credentials"] = "true"

        if self.access_control_expose_headers:
            simple_headers["Access-Control-Expose-Headers"] = ", ".join(self.access_control_expose_headers)

        return simple_headers

    def to_preflight_headers(self):

        allow_all_origins = "*" in self.access_control_allow_origin

        access_control_allow_headers = self.access_control_allow_headers or []
        allow_all_headers = "*" in access_control_allow_headers

        safe_headers = {"Accept", "Accept-Language", "Content-Language", "Content-Type"}


        preflight_explicit_allow_origin = not allow_all_origins or self.access_control_allow_credentials


        preflight_headers: Dict[str, str] = {}
        if preflight_explicit_allow_origin:
            # The origin value will be set in preflight_response() if it is allowed.
            preflight_headers["Vary"] = "Origin"

        else:
            preflight_headers["Access-Control-Allow-Origin"] = "*"

        preflight_headers.update(
            {
                "Access-Control-Allow-Methods": ", ".join(self.access_control_allow_methods),
                "Access-Control-Max-Age": str(self.access_control_max_age),
            }
        )

        allow_headers = sorted(safe_headers | set(access_control_allow_headers))
        
        if allow_headers and not allow_all_headers:
            preflight_headers["Access-Control-Allow-Headers"] = ", ".join(allow_headers)

        if self.access_control_allow_credentials:
            preflight_headers["Access-Control-Allow-Credentials"] = "true"

        return preflight_headers


from enum import Enum


class StageTypes(Enum):
    IDLE='Idle'
    ACT='Act'
    ANALYZE='Analyze'
    EXECUTE='Execute'
    OPTIMIZE='Optimize'
    SETUP='Setup'
    COMPLETE='Complete'
    SUBMIT='Submit'
    ERROR='Error'




import dill
import threading
import os
import pickle
from collections import defaultdict
from typing import Any, Dict, List, Union
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.playwright import (
    MercuryPlaywrightClient, 
    ContextConfig
)
from hedra.core.engines.types.registry import (
    registered_engines,
    RequestTypes
)
from hedra.core.graphs.stages.base.exceptions.process_killed_error import ProcessKilledError
from hedra.core.graphs.stages.base.import_tools import (
    import_stages, 
    import_plugins,
    set_stage_hooks
)
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.execute import Execute
from hedra.core.graphs.stages.optimize.optimization.algorithms import registered_algorithms
from hedra.core.graphs.stages.optimize.optimization import (
    Optimizer, 
    DistributionFitOptimizer
)
from hedra.core.graphs.stages.setup.setup import Setup
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.hooks.types.base.event_graph import EventGraph
from hedra.core.hooks.types.base.registrar import registrar
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.simple_context import SimpleContext
from hedra.core.hooks.types.load.hook import LoadHook
from hedra.core.hooks.types.task.hook import TaskHook
from hedra.core.personas.persona_registry import registered_personas
from hedra.core.personas.streaming.stream_analytics import StreamAnalytics
from hedra.data.serializers import Serializer
from hedra.logging import (
    HedraLogger,
    LoggerTypes
)
from hedra.monitoring import (
    CPUMonitor,
    MemoryMonitor
)
from hedra.plugins.types.plugin_types import PluginType
from hedra.plugins.types.engine.engine_plugin import EnginePlugin
from hedra.plugins.types.persona.persona_plugin import PersonaPlugin
from hedra.versioning.flags.types.base.active import active_flags
from hedra.versioning.flags.types.base.flag_type import FlagTypes


HooksByType = Dict[HookType, Union[List[ActionHook], List[TaskHook]]]


async def setup_action_channels_and_playwright(
    setup_stage: Setup=None,
    execute_stage: Execute=None,
    logger: HedraLogger=None,
    metadata_string: str=None,
    persona_config: Config=None,
    loaded_actions: List[str]=[]
) -> Execute:
    setup_stage.context = SimpleContext()
    setup_stage.generation_setup_candidates = 1
    setup_stage.context['setup_stage_is_primary_thread'] = False
    setup_stage.context['setup_stage_target_config'] = persona_config
    setup_stage.context['setup_stage_target_stages'] = {
        execute_stage.name: execute_stage
    }

    for event in setup_stage.dispatcher.events_by_name.values():
        event.source.stage_instance = setup_stage
        event.context.update(setup_stage.context)

    await setup_stage.run_internal()
    
    stages: Dict[str, Stage] =  setup_stage.context['setup_stage_ready_stages']
    setup_execute_stage: Stage = stages.get(execute_stage.name)
    setup_execute_stage.logger = logger


    actions_and_tasks: List[Union[ActionHook, TaskHook]] = []

    serializer = Serializer()
    if len(loaded_actions) > 0:
        for serialized_action in loaded_actions:
            actions_and_tasks.append(
                serializer.deserialize_action(serialized_action)
            )

    actions_and_tasks.extend(
        setup_stage.context['execute_stage_setup_hooks'].get(execute_stage.name, [])
    )

    for hook in actions_and_tasks:

        if hook.action.type == RequestTypes.PLAYWRIGHT and isinstance(hook.session, MercuryPlaywrightClient):

            await logger.filesystem.aio['hedra.core'].info(f'{metadata_string} - Setting up Playwright Session')

            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Browser Type: {persona_config.browser_type}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Device Type: {persona_config.device_type}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Locale: {persona_config.locale}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - geolocation: {persona_config.geolocation}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Permissions: {persona_config.permissions}')
            await logger.filesystem.aio['hedra.core'].debug(f'{metadata_string} - Playwright Session - {hook.session.session_id} - Color Scheme: {persona_config.color_scheme}')

            config = hook.session.config
            if config is None:
                config = ContextConfig(
                    browser_type=persona_config.browser_type,
                    device_type=persona_config.device_type,
                    locale=persona_config.locale,
                    geolocation=persona_config.geolocation,
                    permissions=persona_config.permissions,
                    color_scheme=persona_config.color_scheme,
                    options=persona_config.playwright_options
                )

            await hook.session.setup(
                config=config
            )

    hooks_by_type: Dict[
        HookType, 
        Union[
            List[ActionHook], 
            List[TaskHook]
        ]
    ] = defaultdict(list)

    for hook in actions_and_tasks:
        hooks_by_type[hook.hook_type].append(hook)

    setup_execute_stage.hooks[HookType.ACTION] = hooks_by_type[HookType.ACTION]
    setup_execute_stage.hooks[HookType.TASK] = hooks_by_type[HookType.TASK]

    for load_hook in setup_execute_stage.hooks[HookType.LOAD]:
        load_hook: LoadHook = load_hook
        load_hook.parser_config = setup_stage.config

    return setup_execute_stage


def optimize_stage(serialized_config: str):

    import asyncio
    import uvloop
    import warnings
    warnings.simplefilter("ignore")

    uvloop.install()

    try:
        loop = asyncio.get_event_loop()
    except Exception:

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    from hedra.logging import (
        logging_manager
    )

    try:

        thread_id = threading.current_thread().ident
        process_id = os.getpid()

        optimization_config: Dict[str, Union[str, int, Any]] = dill.loads(serialized_config)

        graph_name: str = optimization_config.get('graph_name')
        graph_path: str= optimization_config.get('graph_path')
        graph_id: str = optimization_config.get('graph_id')
        logfiles_directory = optimization_config.get('logfiles_directory')
        log_level = optimization_config.get('log_level')
        enable_unstable_features = optimization_config.get('enable_unstable_features', False)
        source_stage_name: str = optimization_config.get('source_stage_name')
        worker_id = optimization_config.get('worker_id')


        monitor_name = f'{source_stage_name}.worker'

        cpu_monitor = CPUMonitor()
        memory_monitor = MemoryMonitor()

        cpu_monitor.stage_type = StageTypes.OPTIMIZE
        memory_monitor.stage_type = StageTypes.OPTIMIZE

        cpu_monitor.start_background_monitor_sync(monitor_name)
        memory_monitor.start_background_monitor_sync(monitor_name)

        active_flags[FlagTypes.UNSTABLE_FEATURE] = enable_unstable_features
    
        logfiles_directory = optimization_config.get('logfiles_directory')
        log_level = optimization_config.get('log_level')

        logging_manager.disable(
            LoggerTypes.DISTRIBUTED,
            LoggerTypes.DISTRIBUTED_FILESYSTEM
        )

        logging_manager.update_log_level(log_level)
        logging_manager.logfiles_directory = logfiles_directory

        logging_manager.disable(
            LoggerTypes.DISTRIBUTED,
            LoggerTypes.DISTRIBUTED_FILESYSTEM,
            LoggerTypes.SPINNER
        )

        logging_manager.update_log_level(log_level)
        logging_manager.logfiles_directory = logfiles_directory

        logger = HedraLogger()
        logger.initialize()
        logger.filesystem.sync.create_logfile('hedra.core.log')
        logger.filesystem.sync.create_logfile('hedra.optimize.log')

        logger.filesystem.create_filelogger('hedra.core.log')
        logger.filesystem.create_filelogger('hedra.optimize.log')
            
        source_stage_id: str = optimization_config.get('source_stage_id')
        source_stage_context: Dict[str, Any] = optimization_config.get('source_stage_context')

        execute_stage_name: str = optimization_config.get('execute_stage_name')
        execute_stage_config: Config = optimization_config.get('execute_stage_config')
        execute_setup_stage_name: Config = optimization_config.get('execute_setup_stage_name')
        execute_stage_plugins: Dict[PluginType, List[str]] = optimization_config.get('execute_stage_plugins')
        execute_stage_streamed_analytics: Dict[str, Dict[str, List[StreamAnalytics]]] = optimization_config.get('execute_stage_streamed_analytics')
        
        optimizer_params: List[str] = optimization_config.get('optimizer_params')
        optimizer_iterations: int = optimization_config.get('optimizer_iterations')
        optimizer_algorithm: str = optimization_config.get('optimizer_algorithm')
        optimize_stage_workers: int = optimization_config.get('optimize_stage_workers')
        time_limit: int = optimization_config.get('time_limit')
        batch_size: int = optimization_config.get('execute_stage_batch_size')
        execute_stage_loaded_actions = optimization_config.get('execute_stage_loaded_actions')

        metadata_string = f'Graph - {graph_name}:{graph_id} - thread:{thread_id} - process:{process_id} - Stage: {source_stage_name}:{source_stage_id} - '
        discovered: Dict[str, Stage] = import_stages(graph_path)
        plugins_by_type = import_plugins(graph_path)

        initialized_stages = {}
        hooks_by_type = defaultdict(dict)
        hooks_by_name = {}
        hooks_by_shortname = defaultdict(dict)

        execute_stage_config.batch_size = batch_size

        generated_hooks = {}
        for stage in discovered.values():
            stage: Stage = stage()
            stage.context = SimpleContext()
            stage.graph_name = graph_name
            stage.graph_path = graph_path
            stage.graph_id = graph_id

            for hook_shortname, hook in registrar.reserved[stage.name].items():
                hook._call = hook._call.__get__(stage, stage.__class__)
                setattr(stage, hook_shortname, hook._call)

            initialized_stage = set_stage_hooks(
                stage, 
                generated_hooks
            )

            for hook_type in initialized_stage.hooks:

                for hook in initialized_stage.hooks[hook_type]:
                    hooks_by_type[hook_type][hook.name] = hook
                    hooks_by_name[hook.name] = hook
                    hooks_by_shortname[hook_type][hook.shortname] = hook

            initialized_stages[initialized_stage.name] = initialized_stage

        execute_stage: Stage = initialized_stages.get(execute_stage_name)
        execute_stage.context.update(source_stage_context)

        setup_stage: Setup = initialized_stages.get(execute_setup_stage_name)
        setup_stage.context.update(source_stage_context)

        stage_persona_plugins: List[str] = execute_stage_plugins[PluginType.PERSONA]
        persona_plugins: Dict[str, PersonaPlugin] = plugins_by_type[PluginType.PERSONA]

        stage_engine_plugins: List[str] = execute_stage_plugins[PluginType.ENGINE]
        engine_plugins: Dict[str, EnginePlugin] = plugins_by_type[PluginType.ENGINE]

        for plugin_name in stage_persona_plugins:
            plugin = persona_plugins.get(plugin_name)
            plugin.name = plugin_name
            registered_personas[plugin_name] = lambda config: plugin(config)
        
        for plugin_name in stage_engine_plugins:
            plugin = engine_plugins.get(plugin_name)
            plugin.name = plugin_name
            registered_engines[plugin_name] = lambda config: plugin(config)

        for plugin_name, plugin in plugins_by_type[PluginType.OPTIMIZER].items():
            registered_algorithms[plugin_name] = plugin

        events_graph = EventGraph(hooks_by_type)
        events_graph.hooks_to_events().assemble_graph().apply_graph_to_events()

        for stage in initialized_stages.values():
            stage.dispatcher.assemble_action_and_task_subgraphs()

        setup_stage: Setup = initialized_stages.get(execute_setup_stage_name)

        setup_execute_stage: Execute = loop.run_until_complete(setup_action_channels_and_playwright(
            setup_stage=setup_stage,
            logger=logger,
            metadata_string=metadata_string,
            persona_config=execute_stage_config,
            execute_stage=execute_stage,
            loaded_actions=execute_stage_loaded_actions
        ))

        pipeline_stages = {
            setup_execute_stage.name: setup_execute_stage
        }

        logger.filesystem.sync['hedra.optimize'].info(f'{metadata_string} - Setting up Optimization')

        if execute_stage_config.experiment and execute_stage_config.experiment.get('distribution'):

            execute_stage_config.experiment['distribution'] = [
                int(distribution_value/optimize_stage_workers) for distribution_value in execute_stage_config.experiment.get('distribution')
            ]

            optimizer = DistributionFitOptimizer({
                'graph_name': graph_name,
                'graph_id': graph_id,
                'source_stage_name': source_stage_name,
                'source_stage_id': source_stage_id,
                'params': optimizer_params,
                'stage_name': execute_stage_name,
                'stage_config': execute_stage_config,
                'stage_hooks': setup_execute_stage.hooks,
                'iterations': optimizer_iterations,
                'algorithm': optimizer_algorithm,
                'time_limit': time_limit,
                'stream_analytics': execute_stage_streamed_analytics
            })

        else:
            optimizer = Optimizer({
                'graph_name': graph_name,
                'graph_id': graph_id,
                'source_stage_name': source_stage_name,
                'source_stage_id': source_stage_id,
                'params': optimizer_params,
                'stage_name': execute_stage_name,
                'stage_config': execute_stage_config,
                'stage_hooks': setup_execute_stage.hooks,
                'iterations': optimizer_iterations,
                'algorithm': optimizer_algorithm,
                'time_limit': time_limit,
                'stream_analytics': execute_stage_streamed_analytics
            })


        results = optimizer.optimize()

        if execute_stage_config.experiment and results.get('optimized_distribution'):
            execute_stage_config.experiment['distribution'] = results.get('optimized_distribution')

        execute_stage_config.batch_size = results.get('optimized_batch_size', execute_stage_config.batch_size)
        execute_stage_config.batch_interval = results.get('optimized_batch_interval', execute_stage_config.batch_gradient)
        execute_stage_config.batch_gradient = results.get('optimized_batch_gradient', execute_stage_config.batch_gradient)

        logger.filesystem.sync['hedra.optimize'].info(f'{optimizer.metadata_string} - Optimization complete')

        context = {}
        for stage in pipeline_stages.values():
            
            stage.context.ignore_serialization_filters = [
                'execute_stage_setup_hooks'
            ]

            for key, value in stage.context.as_serializable():
                try:
                    dill.dumps(value)
                
                except ValueError:
                    stage.context.ignore_serialization_filters.append(key)
                
                except TypeError:
                    stage.context.ignore_serialization_filters.append(key)

                except pickle.PicklingError:
                    stage.context.ignore_serialization_filters.append(key)

            serializable_context = stage.context.as_serializable()
            context.update({
                context_key: context_value for context_key, context_value in serializable_context
            })

        
        cpu_monitor.stop_background_monitor_sync(monitor_name)
        memory_monitor.stop_background_monitor_sync(monitor_name)

        cpu_monitor.close()
        memory_monitor.close()
        
        loop.close()
        
        return dill.dumps({
            'worker_id': worker_id,
            'stage': execute_stage.name,
            'config': execute_stage_config,
            'params': results,
            'context': context,
            'monitoring': {
                'cpu': cpu_monitor.collected,
                'memory': memory_monitor.collected
            }
        })
    
    except KeyboardInterrupt:
        cpu_monitor.stop_background_monitor_sync(monitor_name)
        memory_monitor.stop_background_monitor_sync(monitor_name)

        raise ProcessKilledError()
    
    except BrokenPipeError:
        cpu_monitor.stop_background_monitor_sync(monitor_name)
        memory_monitor.stop_background_monitor_sync(monitor_name)

        raise ProcessKilledError()
    
    except RuntimeError:
        cpu_monitor.stop_background_monitor_sync(monitor_name)
        memory_monitor.stop_background_monitor_sync(monitor_name)

        raise ProcessKilledError()
    
    except Exception as e:
        cpu_monitor.stop_background_monitor_sync(monitor_name)
        memory_monitor.stop_background_monitor_sync(monitor_name)
        
        raise e


import sys
import glob
import uuid
import importlib
import ntpath
import inspect
from pathlib import Path
from typing import List, Union, Dict
from hedra.core.graphs.stages.base.stage import Stage
from hedra.logging import HedraLogger
from hedra.plugins.types.common.plugin import Plugin
from .actions import (
    Fetch,
    Syncrhonize,
    Initialize,
    RepoConfig,
    CreateGitignore
)

from .exceptions import InvalidActionError


class GraphManager:

    def __init__(self, config: RepoConfig, log_level: str='info') -> None:
        self.manager_id = str(uuid.uuid4())
        self._actions = {
            'initialize': Initialize,
            'synchronize': Syncrhonize,
            'fetch': Fetch,
            'create-gitignore': CreateGitignore
        }

        self.discovered_graphs: Dict[str, str] = {}
        self.discovered_plugins: Dict[str, str] = {}
        self.config = config
        self.log_level = log_level
        self.logger = HedraLogger()
        self.logger.initialize()

    def execute_workflow(self, workflow_actions: List[str]):
        
        for workflow_action in workflow_actions:
            self.logger.hedra.sync.debug(f'GraphManager: {self.manager_id} - Executing workflow action - {workflow_action}')

            init_files = glob.glob(self.config.path + '/**/__init__.py', recursive=True)

            discovered = [
                *list(self.discovered_graphs.values()),
                *list(self.discovered_plugins.values()),
                *init_files
            ]
            action: Union[Initialize, Syncrhonize] = self._actions.get(workflow_action)(
                self.config,
                discovered
            )

            if action is None:
                raise InvalidActionError(
                    workflow_action, 
                    list(self._actions.keys())
                )

            action.execute()
            self.logger.hedra.sync.debug(f'GraphManager: {self.manager_id} - Completed workflow action - {workflow_action}')

    def discover_graph_files(self) -> Dict[str, str]:

        candidate_files = glob.glob(self.config.path + '/**/*.py', recursive=True)
        self.logger.hedra.sync.debug(f'GraphManager: {self.manager_id} - Searching for Graph and Plugin files on path - {self.config.path}')

        for candidate_filepath in candidate_files:

            self.logger.hedra.sync.debug(f'GraphManager: {self.manager_id} - Analyzing file: {candidate_filepath}')

            package_dir = Path(candidate_filepath).resolve().parent
            package_dir_path = str(package_dir)
            package_dir_module = package_dir_path.split('/')[-1]
            
            package = ntpath.basename(candidate_filepath)
            package_slug = package.split('.')[0]
            spec = importlib.util.spec_from_file_location(f'{package_dir_module}.{package_slug}', candidate_filepath)

            if candidate_filepath not in sys.path:
                sys.path.append(str(package_dir.parent))

            module = importlib.util.module_from_spec(spec)
            sys.modules[module.__name__] = module

            try:
                spec.loader.exec_module(module)
            
                stage_decendants = list({cls.__name__: cls for cls in Stage.__subclasses__()}.values())
                plugin_decendants = list({cls.__name__: cls for cls in Plugin.__subclasses__()}.values())

                graphs = {}
                plugins = {}
                for name, obj in inspect.getmembers(module):
                    if inspect.isclass(obj) and issubclass(obj, Stage) and obj not in stage_decendants:
                        graphs[name] = obj

                    elif inspect.isclass(obj) and issubclass(obj, Plugin) and obj not in plugin_decendants:
                        plugins[name] = obj


                if len(graphs) > 0:
                    self.logger.hedra.sync.debug(f'GraphManager: {self.manager_id} - Found Graph file at - {candidate_filepath}')               
                    graph_filepath = Path(candidate_filepath)                
                    self.discovered_graphs[graph_filepath.stem] = str(graph_filepath.resolve())

                if len(plugins) > 0:
                    self.logger.hedra.sync.debug(f'GraphManager: {self.manager_id} - Found Plugin file at - {candidate_filepath}')     
                    plugin_filepath = Path(candidate_filepath)
                    self.discovered_plugins[plugin_filepath.stem] = str(plugin_filepath.resolve())

            except Exception as e:
                self.logger.hedra.sync.error(f'Encountered error loading file at - {str(e)}.')
                pass

        
        return {
            'graphs': self.discovered_graphs,
            'plugins': self.discovered_plugins
        }


from typing import List, Tuple, Union
from hedra.distributed.discovery.dns.core.url import (
    URL,
    InvalidHost,
    InvalidIP

)
from hedra.distributed.discovery.dns.core.cache import CacheNode
from hedra.distributed.models.dns import (
    DNSEntry,
    DNSMessage,
    QueryType
)
from hedra.distributed.discovery.dns.core.record import RecordType
from hedra.distributed.discovery.dns.core.record import Record
from hedra.distributed.discovery.dns.core.record.record_data_types import (
    CNAMERecordData,
    NSRecordData
)
from .memoizer import Memoizer


class CacheResolver:
    zone_domains = []
    nameserver_types = [
        RecordType.A
    ]
    memoizer=Memoizer()

    def __init__(    
        self,
        port: int,
        cache: CacheNode = None,
        query_timeout: float = 3.0,
        request_timeout: float = 5.0
    ):
        self.port = port
        self._queries = {}
        self.cache = cache or CacheNode()
        self.request_timeout = request_timeout
        self.query_timeout = query_timeout

    def cache_message(self, entry: DNSEntry):

        for _, record in entry.to_record_data():
            if entry.time_to_live > 0 and record.rtype != RecordType.SOA:
                self.cache.add(record=record)

    def set_zone_domains(self, domains: List[str]):
        self.zone_domains = [
            domain.lstrip('.') for domain in domains
        ]

    async def _query(self, _fqdn: str, _record_type: RecordType) -> Tuple[DNSMessage, bool]:
        raise NotImplementedError

    @memoizer.memoize_async(
            lambda _, fqdn, record_type, skip_cache: (fqdn, record_type)
    )
    async def query_local(
        self,
        fqdn: str,
        record_type: RecordType=RecordType.ANY
    ) -> Tuple[DNSMessage, bool]:
        
        if fqdn.endswith('.'):
            fqdn = fqdn[:-1]

        if record_type == RecordType.ANY:
            try:

                url = URL(
                    fqdn,
                    port=self.port
                )

                ptr_name = url.to_ptr()

            except (InvalidHost, InvalidIP):
                pass

            else:
                fqdn = ptr_name
                record_type = RecordType.PTR

        msg = DNSMessage()
        msg.query_domains.append(
            Record(
                QueryType.REQUEST, 
                name=fqdn, 
                record_type=record_type
            )
        )

        has_result = False
        has_result, fqdn = self.query_cache(msg, fqdn, record_type)

        return msg, has_result

    def _add_cache_cname(self, msg: DNSMessage, fqdn: str) -> Union[str, None]:

        for cname in self.cache.query(fqdn, RecordType.CNAME):

            msg.query_answers.append(cname.copy(name=fqdn))
            if isinstance(cname.data, CNAMERecordData):
                return cname.data.data

    def _add_cache_record_type(
        self, 
        msg: DNSMessage, 
        fqdn: str, 
        record_type: RecordType
    ) -> bool:
        '''Query cache for records other than CNAME and add to result msg.
        '''
        if record_type == RecordType.CNAME:
            return False
        
        has_result = False
        for rec in self.cache.query(fqdn, record_type):

            if isinstance(rec.data, NSRecordData):
                records = list(self.cache.query(
                    rec.data.data, (
                        RecordType.A,
                        RecordType.AAAA
                    )
                ))
                
                if records:
                    msg.query_additional_records.extend(records)
                    msg.query_namservers.append(rec)
                    has_result = True
            else:
                msg.query_answers.append(rec.copy(name=fqdn))
                has_result = True

        return has_result
 
    def query_cache(
        self, 
        fqdn: str, 
        record_type: RecordType
    ) -> Tuple[DNSMessage, bool]:
        
        if fqdn.endswith('.'):
            fqdn = fqdn[:-1]

        if record_type == RecordType.ANY:
            try:

                url = URL(
                    fqdn,
                    port=self.port
                )

                ptr_name = url.to_ptr()

            except (InvalidHost, InvalidIP):
                pass

            else:
                fqdn = ptr_name
                record_type = RecordType.PTR

        msg = DNSMessage()
        msg.query_domains.append(
            Record(
                QueryType.REQUEST, 
                name=fqdn, 
                record_type=record_type
            )
        )

        cnames = set()

        while True:
            cname = self._add_cache_cname(msg, fqdn)
            if not cname: 
                break

            if cname in cnames:
                # CNAME cycle detected
                break

            cnames.add(cname)
            # RFC1034: If a CNAME RR is present at a node, no other data should be present
            fqdn = cname

        has_result = bool(cname) and record_type in (RecordType.CNAME, RecordType.ANY)

        if record_type != RecordType.CNAME:
            has_result = self._add_cache_record_type(
                msg, 
                fqdn, 
                record_type
            ) or has_result

        if any(('.' + fqdn).endswith(root) for root in self.zone_domains):
            if not has_result:
                msg.r = 3
                has_result = True

            msg = DNSMessage(
                **msg.dict(),
                query_authoritative_answer=1
            )
        # fqdn may change due to CNAME
        return msg, has_result



import json
from typing import Any, Union, Dict
from hedra.core.engines.types.udp import UDPResult
from .base_processed_result import BaseProcessedResult


class UDPProcessedResult(BaseProcessedResult):

    __slots__ = (
        'event_id',
        'action_id',
        'url',
        'ip_addr',
        'path',
        'method',
        'headers',
        'params',
        'hostname',
        'status',
        'data',
        'timings'
    )

    def __init__(
        self, 
        stage: str, 
        result: UDPResult
    ) -> None:
        super(UDPProcessedResult, self).__init__(
            stage,
            result
        )

        self.url = result.url
        self.ip_addr = result.ip_addr
        self.path = result.path
        self.method = 'n/a'
        self.headers = {}
        self.params = result.params
        self.hostname = result.hostname
        self.status = result.status
        self.data = result.data
        self.name = self.shortname

        self.time = result.complete - result.start
        
        self.timings = {
            'total': self.time,
            'waiting': result.start - result.wait_start,
            'connecting': result.connect_end - result.start,
            'writing': result.write_end - result.connect_end,
            'reading': result.complete - result.write_end
        }

    def to_dict(self) -> Dict[str, Union[str, int, float]]:

        data = self.data
        if isinstance(data, (bytes, bytearray)):
            data = data.decode()

        return {
            'name': self.name,
            'stage': self.stage,
            'shortname': self.shortname,
            'checks': [check.__name__ for check in self.checks],
            'error': str(self.error),
            'time': self.time,
            'type': self.type,
            'source': self.source,
            'url': self.url,
            'ip_addr': self.ip_addr,
            'method': self.method,
            'path': self.path,
            'params': self.params,
            'hostname': self.hostname,
            'status': self.status,
            'headers': self.headers,
            'data': data
        }

    def serialize(self) -> str:

        data = self.data
        if isinstance(data, (bytes, bytearray)):
            data = data.decode()

        return json.dumps(
            self.to_dict()
        )



import asyncio
import dill
import functools
import signal
import psutil
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from typing import (
    Callable, 
    Awaitable, 
    Any, 
    Optional, 
    Dict, 
    List
)
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.engines.types.common.base_result import BaseResult
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.reporting.metric.custom_metric import CustomMetric


RawResultsSet = Dict[str, ResultsSet]
def handle_loop_stop(
        signame, 
        executor: ThreadPoolExecutor
    ):
    try:
        executor.shutdown(wait=False, cancel_futures=True)

    except BrokenPipeError:
        pass

    except RuntimeError:
        pass

class MetricHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        metric_type: str,
        group: Optional[str] = None,
        order: int=1,
        skip: bool=False
    ) -> None:        
        super().__init__(
            name, 
            shortname, 
            call, 
            order=order,
            skip=skip,
            hook_type=HookType.METRIC
        )
        
        self.names = []
        self.metric_type = metric_type
        self.group = group
        self.order = order
        self.executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=True))
        self.loop: asyncio.AbstractEventLoop = None

    async def call(self, **kwargs):

        if self.skip:
            return kwargs
        
        results: RawResultsSet = self.context['analyze_stage_raw_results']
        self.loop = asyncio.get_running_loop()
        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            self.loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(
                    signame,
                    self.executor
                )
            )


        args_by_hook = {}

        if 'results' in self.params:
            stage_results = await self.loop.run_in_executor(
                self.executor,
                functools.partial(
                    self._generate_deserialized_results,
                    results
                )
            )
            
            for result_set_name, results_set in stage_results.items():
                args_by_hook[result_set_name] =  {
                    **kwargs,
                    'results': results_set
                }

        custom_metric_results: List[Dict[str, CustomMetric]] = await asyncio.gather(*[
            asyncio.create_task(
                self._calculate_stage_custom_metric(
                    args_by_hook[set_name],
                    set_name
                )
            ) for set_name in args_by_hook.keys()
        ])

        metric_results = {}

        for custom_metric in custom_metric_results:
            metric_results.update(custom_metric)

        self.executor.shutdown(wait=False, cancel_futures=True)

        return {
            **kwargs,
            **metric_results
        }
    
    async def _calculate_stage_custom_metric(
            self,
            hook_kwargs: Dict[str, Any],
            set_name: Optional[str]=None
    ) -> Dict[str, CustomMetric]:

        metric_result = await self._call(**{
            name: value for name, value in hook_kwargs.items() if name in self.params
        })
        
        if isinstance(metric_result, dict):
            for metric_name, metric_value in metric_result.items():

                assert isinstance(metric_name, str)
                assert isinstance(metric_value, (float, int))

                metric_fullname = f'{set_name}_{metric_name}'

                custom_metric = CustomMetric(
                    metric_fullname,
                    metric_name,
                    metric_value,
                    metric_group=set_name,
                    metric_type=self.metric_type
                )

                metric_result[metric_name] = custom_metric
                self.context[metric_name] = custom_metric

            return metric_result
        
        assert isinstance(metric_result, (float, int))

        metric_fullname = f'{set_name}_{self.shortname}'
        custom_metric = CustomMetric(
                metric_fullname,
                self.shortname,
                metric_result,
                metric_group=set_name,
                metric_type=self.metric_type
        
        )

        self.context[metric_fullname] = custom_metric
        
        return {
            metric_fullname: custom_metric 
        }

    def _generate_deserialized_results(self, results: RawResultsSet) -> Dict[str, List[BaseResult]]:
        stage_results: Dict[str, List[BaseResult]] = defaultdict(list)

        for results_set in results.values():
            for result in results_set.results:
                stage_result: BaseResult = dill.loads(result)
                stage_results[stage_result.name].append(stage_result)

        return stage_results


    def copy(self):
        metric_hook = MetricHook(
            self.name,
            self.shortname,
            self._call,
            self.group,
            order=self.order,
            skip=self.skip,
        )

        metric_hook.stage = self.stage

        return metric_hook


import asyncio
import math
import os
import psutil
import statistics
from hedra.distributed.models.http import Limit
from typing import Union, List
from .base_limiter import BaseLimiter


class ResourceAdaptiveLimiter(BaseLimiter):

    __slots__ = (
        "max_rate",
        "time_period",
        "_rate_per_sec",
        "_level",
        "_waiters",
        "_loop",
        "_last_check",
        "_cpu_limit",
        "_current_time",
        "_previous_count",
        "_current_cpu",
        "_max_queue",
        "_sample_task",
        "_running",
        "_process",
        "_max_fast_backoff",
        "_min_backoff",
        "_cpu_history",
        "_memory_history",
        "_memory_limit",
        "_current_memory"
    )

    def __init__(
        self,
        limit: Limit
     ) -> None:
        super().__init__(
            limit.max_requests,
            limit.period,
            reject_requests=limit.reject_requests
        )

        cpu_limit = limit.cpu_limit
        if cpu_limit is None:
            cpu_limit = 50

        self._cpu_limit = cpu_limit
        self._backoff = limit.backoff
        self._min_backoff = self._backoff
        self._max_fast_backoff = math.ceil(self._backoff * 10)
        self._max_backoff = math.ceil(self._max_fast_backoff * 10)
        self._last_check = self._loop.time()
        self._current_time = self._loop.time()
        self._previous_count = limit.max_requests

        self._memory_limit = limit.memory

        self._cpu_history: List[float] = []
        self._memory_history: List[float] = []

        self._max_queue = limit.max_requests
        self._sample_task: Union[asyncio.Task, None] = None
        self._running = False
        self._activate_limit = False
        self._process = psutil.Process(os.getpid())

        self._current_cpu = self._process.cpu_percent()
        self._current_memory = self._get_memory()

        self._cpu_history.append(self._current_cpu)
    
    async def acquire(
        self, 
        amount: float = 1,
    ):
        
        if not self._running:
            self._running = True
            self._sample_task = asyncio.create_task(
                self._sample_cpu()
            )
        
        if amount > self.max_rate:
            raise ValueError("Can't acquire more than the maximum capacity")


        task = asyncio.current_task(
            loop=self._loop
        )

        assert task is not None

        rejected = False

        while self._activate_limit:

            fut = self._loop.create_future()
            try:

                self._waiters[task] = fut

                await asyncio.wait_for(
                    asyncio.shield(fut), 
                    timeout=self._backoff
                )

                self._max_fast_backoff = min(
                    self._max_fast_backoff + (1/math.sqrt(self._rate_per_sec)),
                    self._max_backoff
                )

            except asyncio.TimeoutError:
                pass
                
            fut.cancel()
            
            rejected = True

        self._backoff = min(self._backoff * 2, self._max_fast_backoff)
        self._waiters.pop(task, None)
        self._level += amount

        return rejected
    
    async def _sample_cpu(self):
        while  self._running:


            self._current_cpu = self._process.cpu_percent()
            self._current_memory = self._get_memory()

            self._cpu_history.append(self._current_cpu)
            self._memory_history.append(self._current_memory)

            elapsed = self._loop.time() - self._last_check

            if elapsed > self.time_period:
                self._cpu_history.pop(0)


            median_cpu_usage = statistics.median(self._cpu_history)
            median_memory_usage = statistics.median(self._memory_history)


            if self._current_cpu >= self._cpu_limit or self._current_memory >= self._memory_limit:
                self._activate_limit = True

            elif median_cpu_usage < self._cpu_limit and median_memory_usage < self._memory_limit:
                self._activate_limit = False
                self._max_fast_backoff = max(
                    self._max_fast_backoff - (1/self._rate_per_sec),
                    self._min_backoff
                )

            
            await asyncio.sleep(0.1)

    def _get_memory(self):
        return self._process.memory_info().rss / 1024 ** 2

    async def close(self):
        self._running = False

        self._sample_task.cancel()
        if not self._sample_task.cancelled():
            try:

                await self._sample_task

            except (
                asyncio.CancelledError,
                asyncio.InvalidStateError
            ):
                pass


from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class GoogleCloudStorageConfig(BaseModel):
    service_account_json_path: str
    bucket_namespace: str
    events_bucket: str='events'
    metrics_bucket: str='metrics'
    experiments_bucket: str='experiments'
    streams_bucket: str='stages'
    system_metrics_bucket: str='system_metrics'
    reporter_type: ReporterTypes=ReporterTypes.GCS


from hedra.core.graphs.stages import (
    Execute,
)

from hedra.core.hooks import (
    action
)


class ExecuteUDPStage(Execute):

    @action()
    async def udp_send(self):
        return await self.client.udp.send('https://<url_here>', data="PING")

    @action()
    async def udp_receive(self):
        return await self.client.udp.receive('https://<url_here>')


import asyncio
import functools
import psutil
import uuid
from typing import List, Dict
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import (
    MetricsSet,
    MetricType
)
from hedra.reporting.system.system_metrics_set import SystemMetricsSet
from .bigquery_config import BigQueryConfig

try:
    from google.cloud import bigquery
    has_connector = True

except Exception:
    bigquery = None
    has_connector = False


class BigQuery:

    def __init__(self, config: BigQueryConfig) -> None:
        self.service_account_json_path = config.service_account_json_path
        self.project_name = config.project_name
        self.dataset_name = config.dataset_name
        self.retry_timeout = config.retry_timeout

        self.events_table_name = config.events_table

        self.metrics_table_name = config.metrics_table
        self.shared_metrics_table_name = f'{self.metrics_table_name}_shared'
        self.custom_metrics_table_name = f'{self.metrics_table_name}_custom'

        self.stream_metrics_table_name = config.streams_table

        self.errors_table_name = f'{self.metrics_table_name}_errors'
        self.experiments_table_name = config.experiments_table
        self.variants_table_name = f'{self.experiments_table_name}_variants'
        self.mutations_table_name = f'{self.experiments_table_name}_mutations'

        self.session_system_metrics_table_name = f'{config.system_metrics_table}_session'
        self.stage_system_metrics_table_name = f'{config.system_metrics_table}_stage'

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self.credentials = None
        self.client = None

        self._events_table = None
        self._streams_table = None
        self._errors_table = None
        self._metrics_table = None

        self._experiments_table = None
        self._variants_table = None
        self._mutations_table = None

        self._session_system_metrics_table = None
        self._stage_system_metrics_table = None

        self._custom_metrics_table = None
        self._shared_metrics_table = None

        self._loop = asyncio.get_event_loop()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening amd authorizing connection to Google Cloud - Loading account config from - {self.service_account_json_path}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Opening session - {self.session_uuid}')

        self.client = bigquery.Client.from_service_account_json(self.service_account_json_path)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opened connection to Google Cloud - Loaded account config from - {self.service_account_json_path}')

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.session_system_metrics_table_name} - if not exists')
        
        rows = []
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.system_metrics_set_id}')

            if self._session_system_metrics_table is None:

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.session_system_metrics_table_name} - if not exists')

                table_schema = [
                    bigquery.SchemaField('name', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('group', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('median', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('mean', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('variance', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('stdev','FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('minimum', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('maximum', 'FLOAT64', mode='REQUIRED')
                ]

                for quantile in metrics_set.quantiles:
                    table_schema.append(
                        bigquery.SchemaField(quantile, 'FLOAT64')
                    )

                table_reference = bigquery.TableReference(
                    bigquery.DatasetReference(
                        self.project_name,
                        self.dataset_name
                    ),
                    self.session_system_metrics_table_name
                )

                session_system_metrics_table = bigquery.Table(
                    table_reference,
                    schema=table_schema
                )

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.create_table,
                        session_system_metrics_table,
                        exists_ok=True
                    )
                )

                self._session_system_metrics_table = session_system_metrics_table

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.session_system_metrics_table_name} - if not exists')
            
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                rows.append(monitor_metrics.record)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                rows.append(monitor_metrics.record)

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                self._session_system_metrics_table,
                rows,
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.session_system_metrics_table_name} - if not exists')
   
    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.stage_system_metrics_table_name} - if not exists')
        
        rows = []
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.system_metrics_set_id}')

            if self._stage_system_metrics_table is None:

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.stage_system_metrics_table_name} - if not exists')

                table_schema = [
                    bigquery.SchemaField('name', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('stage', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('group', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('median', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('mean', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('variance', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('stdev','FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('minimum', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('maximum', 'FLOAT64', mode='REQUIRED')
                ]

                for quantile in metrics_set.quantiles:
                    table_schema.append(
                        bigquery.SchemaField(quantile, 'FLOAT64')
                    )

                table_reference = bigquery.TableReference(
                    bigquery.DatasetReference(
                        self.project_name,
                        self.dataset_name
                    ),
                    self.stage_system_metrics_table_name
                )

                stage_system_metrics_table = bigquery.Table(
                    table_reference,
                    schema=table_schema
                )

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.create_table,
                        stage_system_metrics_table,
                        exists_ok=True
                    )
                )

                self._stage_system_metrics_table = stage_system_metrics_table

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.stage_system_metrics_table_name} - if not exists')
            
            for metrics_set in system_metrics_sets:

                cpu_metrics = metrics_set.cpu
                memory_metrics = metrics_set.memory

                for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                    for monitor_metrics in stage_cpu_metrics.values():
                        rows.append(monitor_metrics.record)

                    stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                    for monitor_metrics in stage_memory_metrics.values():
                        rows.append(monitor_metrics.record)

                    stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                    
                    if stage_mb_per_vu_metrics:
                        rows.append(stage_mb_per_vu_metrics.record)

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                self._stage_system_metrics_table,
                rows,
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.stage_system_metrics_table_name} - if not exists')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.stream_metrics_table_name} - if not exists')

        rows = []
        for stage_name, stream in stream_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Streams - {stage_name}:{stream.stream_set_id}')

            if self._streams_table is None:

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.stream_metrics_table_name} - if not exists')

                table_schema = [
                    bigquery.SchemaField('name', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('stage', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('group', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('median', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('mean', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('variance', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('stdev','FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('minimum', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('maximum', 'FLOAT64', mode='REQUIRED')
                ]

                for quantile in stream.quantiles:
                    table_schema.append(
                        bigquery.SchemaField(f'{quantile}', 'FLOAT64')
                    )

                table_reference = bigquery.TableReference(
                    bigquery.DatasetReference(
                        self.project_name,
                        self.dataset_name
                    ),
                    self.stream_metrics_table_name
                )

                streams_table = bigquery.Table(
                    table_reference,
                    schema=table_schema
                )

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.create_table,
                        streams_table,
                        exists_ok=True
                    )
                )

                self._streams_table = streams_table

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.stream_metrics_table_name} - if not exists')
            
            for group_name, group in stream.grouped:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Streams Group - {group_name}:{stream.stream_set_id}')
                rows.append({
                    'name': f'{stage_name}_stream',
                    'stage': stage_name,
                    'group': group_name,
                    **group
                })

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                self._streams_table,
                rows,
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.stream_metrics_table_name} - if not exists')

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Experiments to table - {self.experiments_table_name}')

        if self._experiments_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.experiments_table_name} - if not exists')
            
            experiments_table_name = f'{self.project_name}.{self.dataset_name}.{self.experiments_table_name}'

            table_schema = bigquery.Table(
                experiments_table_name,
                schema=[
                    bigquery.SchemaField('experiment_name', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('experiment_randomized', 'BOOLEAN', mode='REQUIRED'),
                    bigquery.SchemaField('experiment_completed', 'INT64', mode='REQUIRED'),
                    bigquery.SchemaField('experiment_succeeded', 'INT64', mode='REQUIRED'),
                    bigquery.SchemaField('experiment_failed', 'INT64', mode='REQUIRED'),
                    bigquery.SchemaField('experiment_median_aps', 'FLOAT64', mode='REQUIRED'),
                ])

            table_reference = bigquery.TableReference(
                bigquery.DatasetReference(
                    self.project_name,
                    self.dataset_name
                ),
                self.experiments_table_name
            )

            self._experiments_table = bigquery.Table(
                table_reference,
                schema=table_schema
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_table,
                    self._experiments_table,
                    exists_ok=True
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.experiments_table_name} - if not exists')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.experiments_table_name}')

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                table_reference,
                experiment_metrics.experiments,
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.experiments_table_name}')

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Variants to table - {self.variants_table_name}')

        if self._variants_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.variants_table_name} - if not exists')
            
            variants_table_name = f'{self.project_name}.{self.dataset_name}.{self.variants_table_name}'

            table_schema = bigquery.Table(
                variants_table_name,
                schema=[
                    bigquery.SchemaField('variant_name', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('variant_experiment', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('variant_weight', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('variant_distribution', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('variant_distribution_interval', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('variant_ratio_completed', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('variant_ratio_succeeded', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('variant_ratio_failed', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('variant_ratio_aps', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('variant_completed', 'INT64', mode='REQUIRED'),
                    bigquery.SchemaField('variant_succeeded', 'INT64', mode='REQUIRED'),
                    bigquery.SchemaField('variant_failed', 'INT64', mode='REQUIRED'),
                    bigquery.SchemaField('variant_actions_per_second', 'FLOAT64', mode='REQUIRED'),
                ])

            table_reference = bigquery.TableReference(
                bigquery.DatasetReference(
                    self.project_name,
                    self.dataset_name
                ),
                self.variants_table_name
            )

            self._variants_table = bigquery.Table(
                table_reference,
                schema=table_schema
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_table,
                    self._variants_table,
                    exists_ok=True
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.variants_table_name} - if not exists')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.variants_table_name}')

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                table_reference,
                experiment_metrics.variants,
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.variants_table_name}')

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Mutations to table - {self.mutations_table_name}')

        if self._mutations_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.mutations_table_name} - if not exists')
            
            mutations_table_name = f'{self.project_name}.{self.dataset_name}.{self.mutations_table_name}'

            table_schema = bigquery.Table(
                mutations_table_name,
                schema=[
                    bigquery.SchemaField('mutation_name', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('mutation_experiment_name', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('mutation_variant_name', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('mutation_chance', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('mutation_targets', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('mutation_type', 'STRING', mode='REQUIRED'),
                ])

            table_reference = bigquery.TableReference(
                bigquery.DatasetReference(
                    self.project_name,
                    self.dataset_name
                ),
                self.mutations_table_name
            )

            self._mutations_table = bigquery.Table(
                table_reference,
                schema=table_schema
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_table,
                    self._mutations_table,
                    exists_ok=True
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.mutations_table_name} - if not exists')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.mutations_table_name}')

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                table_reference,
                experiment_metrics.mutations,
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.mutations_table_name}')

    async def submit_events(self, events: List[BaseProcessedResult]):
        
        if self._events_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.events_table_name} - if not exists')
            
            events_table_name = f'{self.project_name}.{self.dataset_name}.{self.events_table_name}'

            table_schema = bigquery.Table(
                events_table_name,
                schema=[
                    bigquery.SchemaField('name', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('stage', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('time', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('succeeded', 'BOOLEAN', mode='REQUIRED'),
                    bigquery.SchemaField('error', 'STRING')
                ])

            table_reference = bigquery.TableReference(
                bigquery.DatasetReference(
                    self.project_name,
                    self.dataset_name
                ),
                self.events_table_name
            )

            self._events_table = bigquery.Table(
                table_reference,
                schema=table_schema
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_table,
                    self._events_table,
                    exists_ok=True
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.events_table_name} - if not exists')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.events_table_name}')

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                table_reference,
                [event.record for event in events],
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.events_table_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):
        if self.shared_metrics_table_name is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.shared_metrics_table_name} - if not exists')

            table_schema = [
                bigquery.SchemaField('name', 'STRING', mode='REQUIRED'),
                bigquery.SchemaField('stage', 'STRING', mode='REQUIRED'),
                bigquery.SchemaField('group', 'STRING', mode='REQUIRED'),
                bigquery.SchemaField('total', 'INTEGER', mode='REQUIRED'),
                bigquery.SchemaField('succeeded', 'INTEGER', mode='REQUIRED'),
                bigquery.SchemaField('failed', 'INTEGER', mode='REQUIRED'),
                bigquery.SchemaField('actions_per_second', 'FLOAT64', mode='REQUIRED')
            ]

            table_reference = bigquery.TableReference(
                bigquery.DatasetReference(
                    self.project_name,
                    self.dataset_name
                ),
                self.shared_metrics_table_name
            )

            shared_metrics_table = bigquery.Table(
                table_reference,
                schema=table_schema
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_table,
                    shared_metrics_table,
                    exists_ok=True
                )
            )

            self._shared_metrics_table = shared_metrics_table

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.shared_metrics_table_name} - if not exists')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.shared_metrics_table_name} - if not exists')

        rows = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
            rows.append({
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                'group': 'common',
                **metrics_set.common_stats
            })

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                self._shared_metrics_table,
                rows,
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.shared_metrics_table_name} - if not exists')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.metrics_table_name} - if not exists')

        rows = []
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            if self._metrics_table is None:

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.metrics_table_name} - if not exists')

                table_schema = [
                    bigquery.SchemaField('name', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('stage', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('group', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('median', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('mean', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('variance', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('stdev','FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('minimum', 'FLOAT64', mode='REQUIRED'),
                    bigquery.SchemaField('maximum', 'FLOAT64', mode='REQUIRED')
                ]

                for quantile in metrics_set.quantiles:
                    table_schema.append(
                        bigquery.SchemaField(f'{quantile}', 'FLOAT64')
                    )

                for custom_field_name, bigquery_schema_field in metrics_set.custom_schemas.items():
                    table_schema.append(
                        custom_field_name, 
                        bigquery_schema_field
                    ) 

                table_reference = bigquery.TableReference(
                    bigquery.DatasetReference(
                        self.project_name,
                        self.dataset_name
                    ),
                    self.metrics_table_name
                )

                metrics_table = bigquery.Table(
                    table_reference,
                    schema=table_schema
                )

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.create_table,
                        metrics_table,
                        exists_ok=True
                    )
                )

                self._metrics_table = metrics_table

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.metrics_table_name} - if not exists')
            
            for group_name, group in metrics_set.groups.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {group_name}:{group.metrics_group_id}')
                rows.append({
                    'group': group_name,
                    **group.record
                })

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                self._metrics_table,
                rows,
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.metrics_table_name} - if not exists')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics - Project: {self.project_name} - Dataset: {self.dataset_name}')

        if self._custom_metrics_table is None:

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.custom_metrics_table_name} - if not exists')

            table_schema = [
                bigquery.SchemaField('name', 'STRING', mode='REQUIRED'),
                bigquery.SchemaField('stage', 'STRING', mode='REQUIRED'),
                bigquery.SchemaField('group', 'STRING', mode='REQUIRED')
            ]

            for metrics_set in metrics_sets:
                for custom_metric in metrics_set.custom_metrics.values():

                    if custom_metric.metric_type == MetricType.COUNT:
                        table_schema.append(
                            bigquery.SchemaField(
                                custom_metric.metric_name, 
                                'INTEGER', 
                                mode='REQUIRED'
                            )
                        )

                    else:
                        table_schema.append(
                            bigquery.SchemaField(
                                custom_metric.metric_name, 
                                'FLOAT64', 
                                mode='REQUIRED'
                            )
                        )

            table_reference = bigquery.TableReference(
                bigquery.DatasetReference(
                    self.project_name,
                    self.dataset_name
                ),
                self._custom_metrics_table_name
            )

            custom_metrics_table = bigquery.Table(
                table_reference,
                schema=table_schema
            )

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_table,
                    custom_metrics_table,
                    exists_ok=True
                )
            )

            self._custom_metrics_table = custom_metrics_table

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.custom_metrics_table_name} - if not exists')

        rows = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')   
        
            rows.append({
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                'group': 'custom',
                **{
                    custom_metric.metric_name: custom_metric.metric_value for custom_metric in metrics_set.custom_metrics.values()
                }
            })

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                self._custom_metrics_table,
                rows,
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics - Project: {self.project_name} - Dataset: {self.dataset_name}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Errors Metrics Set - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.errors_table_name} - if not exists')

        rows = []
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Errors Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            if self._errors_table is None:

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.errors_table_name} - if not exists')

                table_schema = [
                    bigquery.SchemaField('name', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('stage', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('error_message', 'STRING', mode='REQUIRED'),
                    bigquery.SchemaField('error_count', 'INTEGER', mode='REQUIRED'),
                ]

                table_reference = bigquery.TableReference(
                    bigquery.DatasetReference(
                        self.project_name,
                        self.dataset_name
                    ),
                    self.errors_table_name
                )

                errors_table = bigquery.Table(
                    table_reference,
                    schema=table_schema
                )

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.create_table,
                        errors_table,
                        exists_ok=True
                    )
                )

                self._errors_table = errors_table

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created table - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.errors_table_name} - if not exists')

            rows.extend([
                {
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'error_message': error.get('message'),
                    'error_count': error.get('count')
                } for error in metrics_set.errors
            ])

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.insert_rows_json,
                self._errors_table,
                rows,
                retry=bigquery.DEFAULT_RETRY.with_predicate(
                    lambda exc: exc is not None
                ).with_deadline(
                    self.retry_timeout
                )
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Errors Metrics - Project: {self.project_name} - Dataset: {self.dataset_name} - Table: {self.errors_table_name} - if not exists')

    async def close(self):

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')

        await self._loop.run_in_executor(
            self._executor,
            self.client.close
        )


from __future__ import annotations
from typing import Any, Optional, List, Union, Dict


class SimpleContext:

    def __init__(self, **kwargs) -> None:

        self.known_keys = [
            'stages',
            'visited',
            'results',
            'results_stages',
            'summaries',
            'paths',
            'path_lengths',
            'known_keys'
        ]

        self.ignore_serialization_filters = []

        for kwarg_name, kwarg in kwargs.items():
            object.__setattr__(self, kwarg_name, kwarg)

    def __iter__(self):

        ignore_items = [
            *self.known_keys,
            *self.ignore_serialization_filters
        ]

        for key, value in self.__dict__.items():
            if key.startswith('__') is False and key not in ignore_items:
                yield key, value

    def __getattribute__(self, __name: str) -> Any:
        return object.__getattribute__(self, __name)

    def __setattr__(self, __name: str, __value: Any) -> None:
        object.__setattr__(self, __name, __value)

    def __getitem__(self, name: str) -> Optional[Any]:
        if hasattr(self, name):
            return object.__getattribute__(self, name)

        return None

    def __setitem__(self, name: str, value: Any) -> None:
        object.__setattr__(self, name, value)

    def get(self, name: str) -> Optional[Any]:
        return self.__getitem__(name)       

    def keys(self) -> List[str]: 
        return [key for key in self.__dict__.keys() if key.startswith('__') is False]

    def values(self) -> List[Any]:
        return [value for key, value in self.__dict__.items() if key.startswith('__') is False]
    
    def items(self):
        return [
            (
                key, 
                value
            ) for key, value in self.__dict__.items() if key.startswith('__') is False
        ]

    def remove(self, name: str):
        if name.startswith('__') is False:
            object.__delattr__(self, name)

    def update(self, update_context: Union[SimpleContext, Dict[str, Any]]):
        for context_key, context_value in update_context.items():
            object.__setattr__(self, context_key, context_value)

    def as_serializable(self):

        ignore_items = [
            *self.known_keys,
            *self.ignore_serialization_filters
        ]

        serialization_items = []
        for key, value in self.__dict__.items():
            if key.startswith('__') is False and key not in ignore_items:
                serialization_items.append((
                    key,
                    value
                ))
        
        return serialization_items

    def create_or_update(
        self,
        context_key: str,
        value: Any,
        default: Any
    ):

        if hasattr(self, context_key):
            if isinstance(value, dict):
                exitsting_value: dict = self.__getitem__(context_key)
                exitsting_value.update(value)

                self.__setitem__(context_key, exitsting_value)

            elif isinstance(value, list):
                exitsting_value: list = self.__getitem__(context_key)
                exitsting_value.extend(value)

                self.__setitem__(context_key, exitsting_value)
            
            else:
                self.__setitem__(context_key, value)

        else:
            self.__setitem__(context_key, default)

    def create_if_not_exists(
        self,
        context_key: str,
        value: Any
    ):

        if hasattr(self, context_key) is False:
            self.__setitem__(context_key, value)
    


from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class TelegrafConfig(BaseModel):
    host: str='localhost'
    port: int=8094
    reporter_type: ReporterTypes=ReporterTypes.Telegraf


from __future__ import annotations
import math
from enum import Enum
from typing import Dict, Tuple, Union


class StagePriority(Enum):
    LOW=1
    NORMAL=2
    HIGH=3
    EXCLUSIVE=4
    AUTO=5

    

    @classmethod
    def map(cls, priority: Union[str, None]) -> StagePriority:
        priority_map: Dict[str, StagePriority] = {
            'low': cls.LOW,
            'normal': cls.NORMAL,
            'high': cls.HIGH,
            'exclusive': cls.EXCLUSIVE,
            'auto': cls.AUTO
        }

        if priority is None:
            return StagePriority.AUTO

        return priority_map.get(
            priority,
            StagePriority.AUTO
        )
    
    @classmethod
    def get_worker_allocation_range(
        cls, 
        priority_type: StagePriority,
        pool_size: int
    ) -> Tuple[int, int]:
            

        weight_map: Dict[StagePriority, Tuple[int, int]] = {
            StagePriority.LOW: (
                1, 
                math.ceil(pool_size * 0.25)
            ),
            StagePriority.NORMAL: (
                math.ceil(pool_size * 0.25),
                math.ceil(pool_size * 0.75)
            ),
            StagePriority.HIGH: (
                math.ceil(pool_size * 0.75),
                pool_size
            ),
            StagePriority.EXCLUSIVE: (
                pool_size,
                pool_size,
            ),
            StagePriority.AUTO: (
                1,
                pool_size
            )
        }

        return weight_map.get(priority_type)
    





from decimal import Decimal
from pydantic import (
    BaseModel,
    StrictFloat,
    StrictStr,
    StrictInt
)
from hedra.core.personas.streaming.stream_analytics import StreamAnalytics
from typing import Union, List, Dict


class CommonMetrics(BaseModel):
    total: StrictInt
    succeeded: StrictInt
    failed: StrictInt
    aps: StrictFloat


class TotalMetrics(BaseModel):
    total_med: StrictFloat
    total_μ: StrictFloat
    total_var: StrictFloat
    total_std: StrictFloat
    total_min: StrictFloat
    total_max: StrictFloat
    total_q_10: StrictFloat
    total_q_20: StrictFloat
    total_q_30: StrictFloat
    total_q_40: StrictFloat
    total_q_50: StrictFloat
    total_q_60: StrictFloat
    total_q_70: StrictFloat
    total_q_80: StrictFloat
    total_q_90: StrictFloat
    total_q_95: StrictFloat
    total_q_99: StrictFloat


class WaitingMetrics(TotalMetrics):
    waiting_med: StrictFloat
    waiting_μ: StrictFloat
    waiting_var: StrictFloat
    waiting_std: StrictFloat
    waiting_min: StrictFloat
    waiting_max: StrictFloat
    waiting_q_10: StrictFloat
    waiting_q_20: StrictFloat
    waiting_q_30: StrictFloat
    waiting_q_40: StrictFloat
    waiting_q_50: StrictFloat
    waiting_q_60: StrictFloat
    waiting_q_70: StrictFloat
    waiting_q_80: StrictFloat
    waiting_q_90: StrictFloat
    waiting_q_95: StrictFloat
    waiting_q_99: StrictFloat


class ConnectingMetrics(WaitingMetrics):
    connecting_med: StrictFloat
    connecting_μ: StrictFloat
    connecting_var: StrictFloat
    connecting_std: StrictFloat
    connecting_min: StrictFloat
    connecting_max: StrictFloat
    connecting_q_10: StrictFloat
    connecting_q_20: StrictFloat
    connecting_q_30: StrictFloat
    connecting_q_40: StrictFloat
    connecting_q_50: StrictFloat
    connecting_q_60: StrictFloat
    connecting_q_70: StrictFloat
    connecting_q_80: StrictFloat
    connecting_q_90: StrictFloat
    connecting_q_95: StrictFloat
    connecting_q_99: StrictFloat


class WritingMetrics(ConnectingMetrics):
    writing_med: StrictFloat
    writing_μ: StrictFloat
    writing_var: StrictFloat
    writing_std: StrictFloat
    writing_min: StrictFloat
    writing_max: StrictFloat
    writing_q_10: StrictFloat
    writing_q_20: StrictFloat
    writing_q_30: StrictFloat
    writing_q_40: StrictFloat
    writing_q_50: StrictFloat
    writing_q_60: StrictFloat
    writing_q_70: StrictFloat
    writing_q_80: StrictFloat
    writing_q_90: StrictFloat
    writing_q_95: StrictFloat
    writing_q_99: StrictFloat


class ReadingMetrics(WritingMetrics):
    reading_med: StrictFloat
    reading_μ: StrictFloat
    reading_var: StrictFloat
    reading_std: StrictFloat
    reading_min: StrictFloat
    reading_max: StrictFloat
    reading_q_10: StrictFloat
    reading_q_20: StrictFloat
    reading_q_30: StrictFloat
    reading_q_40: StrictFloat
    reading_q_50: StrictFloat
    reading_q_60: StrictFloat
    reading_q_70: StrictFloat
    reading_q_80: StrictFloat
    reading_q_90: StrictFloat
    reading_q_95: StrictFloat
    reading_q_99: StrictFloat


class GroupMetricsSet(ReadingMetrics):

    def get_group(self, group: str) -> Dict[str, Union[int, float]]:
        return {
            key: value for key, value in self.dict().items() if group in key
        }


    @property
    def total(self) -> Dict[str, Union[int, float]]:
        return {
            key: value for key, value in self.dict().items() if 'total' in key
        }


    @property
    def waiting(self) -> Dict[str, Union[int, float]]:
        return {
            key: value for key, value in self.dict().items() if 'waiting' in key
        }

    @property
    def connecting(self) -> Dict[str, Union[int, float]]:
        return {
            key: value for key, value in self.dict().items() if 'connecting' in key
        }
    
    @property
    def reading(self) -> Dict[str, Union[int, float]]:
        return {
            key: value for key, value in self.dict().items() if 'reading' in key
        }
    
    @property
    def writing(self) -> Dict[str, Union[int, float]]:
        return {
            key: value for key, value in self.dict().items() if 'writing' in key
        }
        

class StageMetrics(BaseModel):
    name: StrictStr
    persona: StrictStr
    batch_size: StrictInt
    total: StrictInt
    succeeded: StrictInt
    failed: StrictInt
    aps: StrictFloat
    time: StrictFloat
    μ_sec: Union[StrictFloat, Decimal]
    μ_waiting: Union[StrictFloat, Decimal]
    μ_connecting: Union[StrictFloat, Decimal]
    μ_reading: Union[StrictFloat, Decimal]
    μ_writing: Union[StrictFloat, Decimal]
    streamed_completion_rates: Union[List[float], None]
    streamed_completed: Union[List[float], None]
    streamed_succeeded: Union[List[float], None]
    streamed_failed: Union[List[float], None]
    streamed_batch_timings: Union[List[float], None]


from enum import Enum


class HookType(Enum):
    ACTION='ACTION'
    CHANNEL='CHANNEL'
    CHECK='CHECK'
    CONTEXT='CONTEXT'
    CONDITION='CONDITION'
    EVENT='EVENT'
    INTERNAL='INTERNAL'
    METRIC='METRIC'
    LOAD='LOAD'
    SAVE='SAVE'
    TASK='TASK'
    TRANSFORM='TRANSFORM'



from __future__ import annotations
from typing import (
    Any, 
    Generic, 
    TypeVar, 
    Dict, 
    List, 
    Union, 
    Optional,
    Tuple
)
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.types.stage_types import StageTypes


T = TypeVar('T')


class BaseEdge(Generic[T]):

    def __init__(self, source: Union[T, Stage], destination: Stage) -> None:
        self.requires = []
        self.provides = []
        
        self.to_stage_names: List[str] = []
        self.from_stage_names: List[str] = []
        self.stages_by_type: Dict[StageTypes, Dict[str, Stage]] = {}
        self.path_lengths: Dict[str, int] = {}
        self.history = {}
        self.edge_data = {}
        self.next_history = {}
        self.visited = []
        self.valid_states = []
        self.descendants = []
        self.all_paths: Dict[str, List[str]] = {}
        self.source = source
        self.destination = destination
        self.timeout = None
        self.folded = False
        self.transition_idx = 0
        self.exception: Optional[Exception] = None

        self.edges_by_name: Dict[str, Stage] = {}

        for stage_type in StageTypes:
            self.stages_by_type[stage_type] = {}

        self.skip_stage = self.source.skip

    def __getitem__(self, key: str):
        return self.history.get(key)


    def __setitem__(self, key: str, value: Any):
        self.history[key] = value
        

    async def transition(self) -> Tuple[None, StageTypes]:
        raise NotImplementedError('Err. - Please implement this method in the Edge class inheriting BaseEdge')

    def update(self, destingation: Stage):
        raise NotImplementedError('Err. - Please implement this method in the Edge class inheriting BaseEdge')

    def split(self) -> None:
        raise NotImplementedError('Err. - Please implement this method in the Edge class inheriting BaseEdge')

    def merge(self) -> None:
        raise NotImplementedError('Err. - Please implement this method in the Edge class inheriting BaseEdge')
    
    def setup(self) -> None:
        pass
        #raise NotImplementedError('Err. - Please implement this method in the Edge class inheriting BaseEdge')


import asyncio
import signal
from asyncio import (
    Transport,
    DatagramTransport,
    Protocol,
    sleep
)


class Writer:
    __slots__ = (
        '_transport',
        '_protocol',
        '_reader',
        '_loop',
        '_complete_fut'
    )
    """Wraps a Transport.
    This exposes write(), writelines(), [can_]write_eof(),
    get_extra_info() and close().  It adds drain() which returns an
    optional Future on which you can wait for flow control.  It also
    adds a transport property which references the Transport
    directly.
    """

    def __init__(self, transport, protocol, reader, loop):
        self._transport: Transport | DatagramTransport = transport
        self._protocol: Protocol = protocol
        # drain() expects that the reader has an exception() method
        self._reader = reader
        self._loop: asyncio.AbstractEventLoop = loop
        self._complete_fut = self._loop.create_future()
        self._complete_fut.set_result(None)

    def __repr__(self):
        info = [self.__class__.__name__, f'transport={self._transport!r}']
        if self._reader is not None:
            info.append(f'reader={self._reader!r}')
        return '<{}>'.format(' '.join(info))

    @property
    def transport(self):
        return self._transport

    def write(self, data):
        self._transport.write(data)

    def send(self, data):
        self._transport.sendto(data)

    def writelines(self, data):
        self._transport.writelines(data)

    def write_eof(self):
        return self._transport.write_eof()

    def can_write_eof(self):
        return self._transport.can_write_eof()

    def close(self):
        return self._transport.close()

    def is_closing(self):
        return self._transport.is_closing()

    async def wait_closed(self):
        await self._protocol._get_close_waiter(self)

    def get_extra_info(self, name, default=None):
        return self._transport.get_extra_info(name, default)

    async def drain(self):
        """Flush the write buffer.
        The intended use is to write
          w.write(data)
          await w.drain()
        """
        if self._reader is not None:
            exc = self._reader.exception()
            if exc is not None:
                raise exc
        if self._transport.is_closing():
            # Wait for protocol.connection_lost() call
            # Raise connection closing error if any,
            # ConnectionResetError otherwise
            # Yield to the event loop so connection_lost() may be
            # called.  Without this, _drain_helper() would return
            # immediately, and code that calls
            #     write(...); await drain()
            # in a loop would never call connection_lost(), so it
            # would not see an error when the socket is closed.
            await sleep(0)
        await self._protocol._drain_helper()

    async def start_tls(self, sslcontext, *, server_hostname=None,  ssl_handshake_timeout=None):

        server_side = self._protocol._client_connected_cb is not None
        protocol = self._protocol

        await self.drain()

        new_transport = await self._loop.start_tls(  # type: ignore
            self._transport, 
            protocol, 
            sslcontext,
            server_side=server_side, 
            server_hostname=server_hostname,
            ssl_handshake_timeout=ssl_handshake_timeout
        )
        
        self._transport = new_transport
        protocol._replace_writer(self)


import asyncio
import time
import uuid
from collections import deque
from typing import Dict, Any, Union, Coroutine, TypeVar, Optional
from hedra.core.engines.types.common.base_engine import BaseEngine
from hedra.core.engines.types.common.ssl import get_default_ssl_context
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.common.concurrency import Semaphore
from hedra.core.engines.types.common.protocols.udp.quic_protocol import (
    HeadersState,
    ResponseFrameCollection,
    FrameType,
    encode_frame
)
from hedra.core.engines.types.tracing.trace_session import (
    TraceSession, 
    Trace
)

from hedra.logging import HedraLogger
from hedra.versioning.flags.types.unstable.flag import unstable
from .connection import HTTP3Connection
from .action import HTTP3Action
from .result import HTTP3Result
from .pool import Pool


A = TypeVar('A')
R = TypeVar('R')


@unstable
class MercuryHTTP3Client(BaseEngine[Union[A, HTTP3Action], Union[R, HTTP3Result]]):

    __slots__ = (
        'session_id',
        'timeouts',
        'registered',
        '_hosts',
        'closed',
        'sem',
        'pool',
        'active',
        'waiter',
        'ssl_context',
        'logger',
        'tracing_session'
    )

    def __init__(
        self, 
        concurrency: int=10**3, 
        timeouts: Timeouts = Timeouts(), 
        reset_connections: bool=False,
        tracing_session: Optional[TraceSession]=None
    ) -> None:
        super().__init__()

        self.session_id = str(uuid.uuid4())
        self.timeouts = timeouts

        self.registered: Dict[str, HTTP3Action] = {}
        self._hosts = {}
        self.closed = False

        self.sem = asyncio.Semaphore(value=concurrency)
        self.pool = Pool(concurrency, reset_connections=reset_connections)
        self.tracing_session: Union[TraceSession, None] = tracing_session
        self.logger = HedraLogger()
        self.logger.initialize()
        self.pool.create_pool()
        self.active = 0
        self.waiter = None

        self.ssl_context = get_default_ssl_context()
    
    def config_to_dict(self):
        return {
            'concurrency': self.pool.size,
            'timeouts': {
                'connect_timeout': self.timeouts.connect_timeout,
                'socket_read_timeout': self.timeouts.socket_read_timeout,
                'total_timeout': self.timeouts.total_timeout
            },
            'reset_connections': self.pool.reset_connections
        }


    async def set_pool(self, concurrency: int):
        self.sem = asyncio.Semaphore(value=concurrency)
        self.pool = Pool(concurrency, reset_connections=self.pool.reset_connections)
        self.pool.create_pool()

    def extend_pool(self, increased_capacity: int):
        self.pool.size += increased_capacity
        for _ in range(increased_capacity):
            self.pool.connections.append(
                HTTP3Connection(self.pool.reset_connections)
            )
        
        self.sem = Semaphore(self.pool.size)

    def shrink_pool(self, decrease_capacity: int):
        self.pool.size -= decrease_capacity
        self.pool.connections = self.pool.connections[:self.pool.size]
        self.sem = Semaphore(self.pool.size)
    
    async def prepare(self, action: HTTP3Action) -> Coroutine[Any, Any, None]:
        try:
                
            if self._hosts.get(action.url.hostname) is None:
                socket_configs = await asyncio.wait_for(action.url.lookup(), timeout=self.timeouts.connect_timeout)
            
                for ip_addr, configs in socket_configs.items():
                    for config in configs:

                        connection = HTTP3Connection()
                        
                        try:
                            await connection.make_connection(
                                ip_addr,
                                action.url.port,
                                config,
                                server_name=action.url.hostname,
                                timeout=self.timeouts.connect_timeout
                            )

                            action.url.socket_config = config
                            action.url.ip_addr = ip_addr
                            action.url.has_ip_addr = True
                            break

                        except Exception as e:
                            pass

                    if action.url.socket_config:
                        break

                if action.url.socket_config is None:
                    raise Exception('Err. - No socket found.')

                self._hosts[action.url.hostname] = {
                    'ip_addr': action.url.ip_addr,
                    'socket_config': action.url.socket_config
                }

            else:
                host_config = self._hosts[action.url.hostname]
                action.url.ip_addr = host_config.get('ip_addr')
                action.url.socket_config = host_config.get('socket_config')

            if action.is_setup is False:
                action.setup()

            self.registered[action.name] = action

        except Exception as e:       
            raise e

    async def execute_prepared_request(self, action: HTTP3Action) -> Coroutine[Any, Any, HTTP3Result]:

        trace: Union[Trace, None] = None
        if self.tracing_session:
            trace = self.tracing_session.create_trace()
            await trace.on_request_start(action)
  
        response = HTTP3Result(action)
        response.wait_start = time.monotonic()
        self.active += 1

        if trace and trace.on_connection_queued_start:
            await trace.on_connection_queued_start(
                trace.span,
                action,
                response
            )
 
        async with self.sem:

            if trace and trace.on_connection_queued_end:
                await trace.on_connection_queued_end(
                    trace.span,
                    action,
                    response
                )
            
            try:

                connection = self.pool.connections.pop()
                
                if action.hooks.listen:
                    event = asyncio.Event()
                    action.hooks.channel_events.append(event)
                    await event.wait()

                if action.hooks.before:
                    action = await self.execute_before(action)
                    action.setup()

                response.start = time.monotonic()

                if trace and trace.on_connection_create_start:
                    await trace.on_connection_create_start(
                        trace.span,
                        action,
                        response
                    )

                await connection.make_connection(
                    action.url.ip_addr,
                    action.url.port,
                    action.url.socket_config,
                    server_name=action.url.hostname,
                    timeout=self.timeouts.connect_timeout
                )

                response.connect_end = time.monotonic()

                if trace and trace.on_connection_create_end:
                    await trace.on_connection_create_end(
                        trace.span,
                        action,
                        response
                    )

                stream_id = connection.protocol._quic.get_next_available_stream_id()
                
                stream = connection.protocol._get_or_create_stream(stream_id)
                if stream.headers_send_state == HeadersState.AFTER_TRAILERS:
                    raise Exception("HEADERS frame is not allowed in this state")

                encoder, frame_data = connection.protocol._encoder.encode(
                    stream_id, 
                    action.encoded_headers
                )

                connection.protocol._encoder_bytes_sent += len(encoder)
                connection.protocol._quic.send_stream_data(
                    connection.protocol._local_encoder_stream_id, 
                    encoder
                )

                # update state and send headers
                if stream.headers_send_state == HeadersState.INITIAL:
                    stream.headers_send_state = HeadersState.AFTER_HEADERS
                else:
                    stream.headers_send_state = HeadersState.AFTER_TRAILERS

                connection.protocol._quic.send_stream_data(
                    stream_id, encode_frame(
                        FrameType.HEADERS, 
                        frame_data
                    ), 
                    not action.encoded_data
                )

                if trace and trace.on_request_headers_sent:
                    await trace.on_request_headers_sent(
                        trace.span,
                        action,
                        response
                    )

                if action.encoded_data:
                    stream = connection.protocol._get_or_create_stream(stream_id)
                    if stream.headers_send_state != HeadersState.AFTER_HEADERS:
                        raise Exception("DATA frame is not allowed in this state")

                    connection.protocol._quic.send_stream_data(
                        stream_id, encode_frame(
                            FrameType.DATA, 
                            action.encoded_data
                        ), 
                        True
                    )
                    

                waiter = connection.protocol._loop.create_future()
                connection.protocol._request_events[stream_id] = deque()
                connection.protocol._request_waiter[stream_id] = waiter
                connection.protocol.transmit()

                response.write_end = time.monotonic()

                if action.encoded_data and trace and trace.on_request_data_sent:
                        await trace.on_request_data_sent(
                            trace.span,
                            action,
                            response
                        )

                response_frames: ResponseFrameCollection = await asyncio.wait_for(
                    waiter,
                    timeout=self.timeouts.total_timeout
                )

                headers: Dict[str, Union[bytes, int]] = {}
                for header_key, header_value in response_frames.headers_frame.headers:
                    headers[header_key] = header_value

                if trace and trace.on_response_headers_received:
                    await trace.on_response_headers_received(
                        trace.span,
                        action,
                        response
                    )
                
                response.headers = headers
                status = response.status

                if status >= 300 and status < 400:

                    if trace and trace.on_request_redirect:
                        await trace.on_request_redirect(
                            trace.span,
                            action,
                            response
                        )
                    
                    elapsed_time = 0
                    redirect_time_start = time.time()

                    for redirect_number in range(action.redirects):

                        if elapsed_time > self.timeouts.total_timeout:
                            response.status = 408
                            raise Exception('Request timed out while redirecting.')
                        
                        redirect_url = str(headers.get(b'location'))
                        if redirect_url.startswith('http') is False:
                            action.url.path = redirect_url
                            action.encoded_headers = None
                            action.setup()

                        else:
                            await action.url.replace(redirect_url)
                            action.encoded_headers = None
                            action.is_setup = False
                            await self.prepare(action)

                            await connection.make_connection(
                                action.url.ip_addr,
                                action.url.port,
                                action.url.socket_config,
                                server_name=action.url.hostname,
                                timeout=self.timeouts.connect_timeout
                            )

                        response.connect_end = time.monotonic()
                            
                        stream_id = connection.protocol._quic.get_next_available_stream_id()
                        
                        stream = connection.protocol._get_or_create_stream(stream_id)
                        if stream.headers_send_state == HeadersState.AFTER_TRAILERS:
                            raise Exception("HEADERS frame is not allowed in this state")

                        encoder, frame_data = connection.protocol._encoder.encode(
                            stream_id, 
                            action.encoded_headers
                        )

                        connection.protocol._encoder_bytes_sent += len(encoder)
                        connection.protocol._quic.send_stream_data(
                            connection.protocol._local_encoder_stream_id, 
                            encoder
                        )

                        # update state and send headers
                        if stream.headers_send_state == HeadersState.INITIAL:
                            stream.headers_send_state = HeadersState.AFTER_HEADERS
                        else:
                            stream.headers_send_state = HeadersState.AFTER_TRAILERS
                        connection.protocol._quic.send_stream_data(
                            stream_id, encode_frame(
                                FrameType.HEADERS, 
                                frame_data
                            ), 
                            not action.encoded_data
                        )

                        if action.encoded_data:
                            stream = connection.protocol._get_or_create_stream(stream_id)
                            if stream.headers_send_state != HeadersState.AFTER_HEADERS:
                                raise Exception("DATA frame is not allowed in this state")

                            connection.protocol._quic.send_stream_data(
                                stream_id, encode_frame(
                                    FrameType.DATA, 
                                    action.encoded_data
                                ), 
                                True
                            )
                            

                        waiter = connection.protocol._loop.create_future()
                        connection.protocol._request_events[stream_id] = deque()
                        connection.protocol._request_waiter[stream_id] = waiter
                        connection.protocol.transmit()

                        response.write_end = time.monotonic()

                        response_frames: ResponseFrameCollection = await asyncio.wait_for(
                            waiter,
                            timeout=self.timeouts.total_timeout
                        )

                        for header_key, header_value in response_frames.headers_frame.headers:
                            headers[header_key] = header_value

                        response.headers = headers
                        status = response.status
                        if status >= 200 and status < 300:
                            break

                        action.redirects -= redirect_number
                        elapsed_time = time.time() - redirect_time_start

                response.complete = time.monotonic()

                if trace and trace.on_response_data_received:
                    await trace.on_response_data_received(
                        trace.span,
                        action,
                        response
                    )

                response.headers = headers
                self.pool.connections.append(connection)

                if action.hooks.after:
                    response = await self.execute_after(action, response)
                    action.setup()

                if action.hooks.checks:
                    response = await self.execute_checks(action, response)

                if action.hooks.notify:
                    await asyncio.gather(*[
                        asyncio.create_task(
                            channel.call(response, action.hooks.listeners)
                        ) for channel in action.hooks.channels
                    ])

                    for listener in action.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            listener.setup()
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()       

            except Exception as e:
                response.complete = time.monotonic()
                response.error = str(e)

                self.pool.connections.append(HTTP3Connection(reset_connection=self.pool.reset_connections))

                if trace and trace.on_request_exception:
                    await trace.on_request_exception(response)

            self.active -= 1
            if self.waiter and self.active <= self.pool.size:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None

            if trace and trace.on_request_end:
                await trace.on_request_end(response)

            return response

    async def close(self):
        if self.closed is False:
            await self.pool.close()
            self.closed = True
